from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionFeatureExtraction


class AlicevisionFeatureExtraction(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionFeatureExtraction, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Input Images Directory', NodeParameterType.STRING, "`task['av_camera_sfm']`")
            ui.add_parameter('output', 'Output Feature Dir', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/FeatureExtraction")
            ui.add_separator()
            ui.add_parameter('describerTypes', 'Describer Types', NodeParameterType.STRING, 'sift')  # TODO: make a set of checkboxes
            ui.add_parameter('describerPreset', 'Describer Density', NodeParameterType.STRING, 'normal')\
                .add_menu((('low', 'low'), ('medium', 'medium'), ('normal', 'normal'), ('high', 'high'), ('ultra', 'ultra')))
            ui.add_parameter('describerQuality', 'Describer Quality', NodeParameterType.STRING, 'normal')\
                .add_menu((('low', 'low'), ('medium', 'medium'), ('normal', 'normal'), ('high', 'high'), ('ultra', 'ultra')))
            ui.add_parameter('forceCpuExtraction', 'Force Cpu', NodeParameterType.BOOL, True)
            # TODO: --contrastFiltering, --gridFiltering

            ui.add_separator()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('rangeStart', 'Image Range Start', NodeParameterType.INT, 0)
                ui.add_parameter('rangeSize', 'Image Chunk Size', NodeParameterType.INT, 999999)


    @classmethod
    def label(cls) -> str:
        return 'AV Feature Extraction'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'feature', 'extraction'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_feature_extraction'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output'))
        srcpath = Path(context.param_value('input'))
        if not Path(srcpath).is_absolute():
            raise ProcessingError('image directory path must be absolute and not empty')
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker

        args = ['aliceVision_featureExtraction', '--verboseLevel', 'info']
        args += ['--describerTypes', context.param_value('describerTypes')]
        args += ['--describerPreset', context.param_value('describerPreset')]
        args += ['--describerQuality', context.param_value('describerQuality')]
        args += ['--forceCpuExtraction', context.param_value('forceCpuExtraction')]

        args += ['--rangeStart', context.param_value('rangeStart')]
        args += ['--rangeSize', context.param_value('rangeSize')]

        args += ['--input', srcpath]
        args += ['--output', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_features_dir', str(output))
        return res



from lifeblood.basenode import BaseNode, ProcessingError
from lifeblood.nodethings import ProcessingResult
from lifeblood.processingcontext import ProcessingContext
from lifeblood.enums import NodeParameterType
from lifeblood.environment_resolver import EnvironmentResolverArguments
from lifeblood.uidata import NodeUi, MultiGroupLayout, Parameter
from lifeblood.node_visualization_classes import NodeColorScheme

from typing import Iterable


def node_class():
    return EnvironmentResolverArgumentsSetter


class EnvironmentResolverArgumentsSetter(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'modify environment resolver arguments'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'modify', 'environment', 'resolver'

    @classmethod
    def type_name(cls) -> str:
        return 'environment_resolver_arguments_setter'

    def __init__(self, name: str):
        super(EnvironmentResolverArgumentsSetter, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            mode_param = ui.add_parameter('mode', 'mode', NodeParameterType.STRING, 'modify')\
                                .add_menu((('Set', 'set'),
                                           ('Modify', 'modify')))
            with ui.parameters_on_same_line_block():
                ui.add_parameter('set resolver name', 'set resolver', NodeParameterType.BOOL, False).append_visibility_condition(mode_param, '==', 'modify')
                ui.add_parameter('resolver name', 'resolver', NodeParameterType.STRING, '')
            ui.add_separator()
            with ui.multigroup_parameter_block('arguments', 'arguments to set'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('arg name', '<name', NodeParameterType.STRING, 'arg name')
                    type_param = ui.add_parameter('type', 'value>', NodeParameterType.INT, 0)
                    type_param.add_menu((('int', NodeParameterType.INT.value),
                                         ('bool', NodeParameterType.BOOL.value),
                                         ('float', NodeParameterType.FLOAT.value),
                                         ('string', NodeParameterType.STRING.value)))

                    ui.add_parameter('svalue', 'val', NodeParameterType.STRING, '').append_visibility_condition(type_param, '==', NodeParameterType.STRING.value)
                    ui.add_parameter('ivalue', 'val', NodeParameterType.INT, 0).append_visibility_condition(type_param, '==', NodeParameterType.INT.value)
                    ui.add_parameter('fvalue', 'val', NodeParameterType.FLOAT, 0.0).append_visibility_condition(type_param, '==', NodeParameterType.FLOAT.value)
                    ui.add_parameter('bvalue', 'val', NodeParameterType.BOOL, False).append_visibility_condition(type_param, '==', NodeParameterType.BOOL.value)

            ui.add_separator()
            with ui.multigroup_parameter_block('arguments to delete', 'arguments to delete'):
                ui.add_parameter('arg name to delete', 'argument to delete', NodeParameterType.STRING, '')

            ui.parameter('arguments to delete').append_visibility_condition(mode_param, '==', 'modify')

        # now initialize default values
        ui.parameter('arguments').set_value(1)
        ui.parameter('arg name_0').set_value('user')
        ui.parameter('type_0').set_value(NodeParameterType.STRING.value)
        ui.parameter('svalue_0').set_value('someusername')

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        resolver_name = context.param_value('resolver name').strip()
        arguments = {}
        for i in range(context.param_value('arguments')):
            # we are being sane with names, though name can technically be any unicode string
            name = context.param_value(f'arg name_{i}').strip()
            if name == '':
                continue

            argtype = context.param_value(f'type_{i}')
            if argtype == NodeParameterType.INT.value:
                val = context.param_value(f'ivalue_{i}')
            elif argtype == NodeParameterType.BOOL.value:
                val = context.param_value(f'bvalue_{i}')
            elif argtype == NodeParameterType.FLOAT.value:
                val = context.param_value(f'fvalue_{i}')
            elif argtype == NodeParameterType.STRING.value:
                val = context.param_value(f'svalue_{i}')
            else:
                raise ProcessingError(f'unknown argument type: {argtype}')
            arguments[name] = val

        if context.param_value('mode') == 'set':
            res = ProcessingResult()
            res.set_environment_resolver_arguments(EnvironmentResolverArguments(resolver_name, arguments))

            return res
        else:  # mode is modify
            if not context.param_value('set resolver name'):
                resolver_name = None

        # else mode is modify (no other modes exist)
        env_arguments = context.task_environment_resolver_arguments()
        if resolver_name is not None:
            env_arguments.set_name(resolver_name)
        for arg, val in arguments.items():
            env_arguments.add_argument(arg, val)

        for i in range(context.param_value('arguments to delete')):
            name = context.param_value(f'arg name to delete_{i}').strip()
            if name == '':
                continue
            if name in env_arguments.arguments():
                env_arguments.remove_argument(name)

        res = ProcessingResult()
        res.set_environment_resolver_arguments(env_arguments)

        return res




# NOTE: this file is a simplified standalone version of lifeblood.environment_resolver
# And therefore this file should be kept up to date with the original
#
# note as well: this is still kept py2-3 compatible

import json


class EnvironmentResolverArguments:
    """
    this class objects specity requirements a task/invocation have for int's worker environment wrapper.
    """
    def __init__(self, resolver_name=None, arguments=None):
        """

        :param resolver_name: if None - treat as no arguments at all
        :param arguments:
        """
        if arguments is None:
            arguments = {}
        if resolver_name is None and len(arguments) > 0:
            raise ValueError('if name is None - no arguments are allowed')
        self.__resolver_name = resolver_name
        self.__args = arguments

    def name(self):
        return self.__resolver_name

    def arguments(self):
        return self.__args

    def serialize(self):  # type: () -> bytes
        return json.dumps(self.__dict__).encode('utf-8')

    @classmethod
    def deserialize(cls, data):  # type: (bytes) -> EnvironmentResolverArguments
        wrp = EnvironmentResolverArguments(None)
        wrp.__dict__.update(json.loads(data.decode('utf-8')))
        return wrp



from .messages import Message
from .client import MessageClient


class MessageHandlerBase:
    async def process_message(self, message: Message, client: MessageClient) -> bool:
        """
        should return True if message was processed.
        if message is not expected by this handler - the function should return False
        """
        raise NotImplementedError()



import os
from math import sqrt, floor, ceil

from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.uidata import NodeParameterType
from lifeblood.invocationjob import InvocationJob
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


def node_class():
    return Ffmpeg


class Ffmpeg(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'ffmpeg'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'ffmpeg', 'video', 'convert'

    @classmethod
    def type_name(cls) -> str:
        return 'ffmpeg'

    def __init__(self, name):
        super(Ffmpeg, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.125, 0.33, 0.05)
            ui.add_parameter('sequence', 'sequence attribute name', NodeParameterType.STRING, 'sequence')
            ui.add_parameter('ffmpeg bin path', 'ffmpeg binary', NodeParameterType.STRING, 'ffmpeg')
            ui.add_parameter('fps', 'fps', NodeParameterType.INT, 24)
            ui.add_parameter('icrf', 'quality %', NodeParameterType.INT, 75).set_slider_visualization(1, 100)
            ui.add_parameter('-pix_fmt', 'pixel format', NodeParameterType.STRING, 'yuv420p')
            ui.add_parameter('-vcodec', 'codec', NodeParameterType.STRING, 'libx264')
            ui.add_parameter('outpath', 'movie path', NodeParameterType.STRING, '/tmp/movie.mp4')
            docc = ui.add_parameter('docc', 'color correct', NodeParameterType.BOOL, False)
            ui.add_parameter('cc gamma', 'gamma', NodeParameterType.FLOAT, 1.0).append_visibility_condition(docc, '==', True)

    def process_task(self, context) -> ProcessingResult:
        binpath = context.param_value('ffmpeg bin path')
        attributes = context.task_attributes()
        sequence_attrib_name = context.param_value('sequence')
        if sequence_attrib_name not in attributes:
            raise ProcessingError(f'required attribute "{sequence_attrib_name}" not found')
        sequence = attributes[sequence_attrib_name]

        assert isinstance(sequence, list), 'sequence attribute is supposed to be list of frames, or list of sequences'

        outopts = ['-vcodec', context.param_value('-vcodec'),
                   '-crf', 100 - context.param_value('icrf'),
                   '-pix_fmt', context.param_value('-pix_fmt')]
        fps = context.param_value('fps')
        outpath = context.param_value('outpath')

        # TODO: MOVE THIS TO WORKER! AND MIND POTENTIAL EMBEDDED FILES IN FUTURE
        os.makedirs(os.path.dirname(outpath), exist_ok=True)

        if isinstance(sequence[0], str):  # we have sequence of frames
            filterlines = []
            filterlines.append(f"color=c='Black':r={fps}[null];"
                               f'[null][0]scale2ref=w=iw:h=ih[prebase][sink];[sink]nullsink;[prebase]setsar=1/1[base0];'
                               f'[base0][0]overlay=shortest=1')

            if context.param_value('docc'):
                gam = context.param_value("cc gamma")
                filterlines.append(f'lutrgb=r=gammaval({gam}):g=gammaval({gam}):b=gammaval({gam})')  # this does NOT work for float pixels tho. seems that nothing in FFMPEG does
            args = [binpath, '-f', 'concat', '-safe', '0', '-r', fps, '-i', ':/framelist.txt']
            args += ['-filter_complex', ','.join(filterlines)]  # , here - cuz we just chain those guys together. ";" is for separating chains, not filters in single chain as it is here
            args += [*outopts, '-y', outpath]
            job = InvocationJob(args)
            framelist = '\n'.join(f'file \'{seqitem}\'' for seqitem in sequence)  # file in ffmpeg concat format
            job.set_extra_file('framelist.txt', framelist)
        elif isinstance(sequence[0], list):  # WTF is this? :) i forgot... but #TODO: black base for image needed in case above too
            # for now logic is this:
            # calc dims
            # pick first sequence, use as size ref with dims
            # woop woop
            num_items = len(sequence)
            count_h = ceil(sqrt(num_items))
            count_w = ceil(num_items/count_h)

            filter = f"color=c='Black':r={fps}[null];" \
                     f'[null][0]scale2ref=w=iw*{count_w}:h=ih*{count_h}[prebase][sink];[sink]nullsink;[prebase]setsar=1/1[base0];'

            filterlines = []
            input = 0
            for h in range(count_h):
                for w in range(count_w):
                    filterlines.append(f"[base{input}][{input}]overlay={'shortest=1:' if input == 0 else ''}x='w*{w}':y='h*{h}'[base{input+1}]")
                    input += 1
                    if input == num_items:
                        break
                if input == num_items:
                    break
            if context.param_value('docc'):
                gam = context.param_value("cc gamma")
                filterlines.append(f'[base{input}]lutrgb=r=gammaval({gam}):g=gammaval({gam}):b=gammaval({gam})')  # this does NOT work for float pixels tho. seems that nothing in FFMPEG does
            else:
                filterlines.append(f'[base{input}]null')
            filter += ';'.join(filterlines)

            args = [binpath]
            for i in range(num_items):
                args += ['-f', 'concat', '-safe', '0', '-r', fps, '-i', f':/framelist{i}.txt']
            args += ['-filter_complex', filter]
            args += [*outopts, '-y', outpath]
            job = InvocationJob(args)
            for i, subsequence in enumerate(sequence):
                framelist = '\n'.join(f'file \'{seqitem}\'' for seqitem in subsequence)  # file in ffmpeg concat format
                job.set_extra_file(f'framelist{i}.txt', framelist)
        else:
            raise RuntimeError('bad attribute value')

        res = ProcessingResult(job)
        res.set_attribute('file', outpath)
        return res


    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import struct
import asyncio
import aiofiles
from enum import Enum
import pickle
import json

from . import logging
from . import invocationjob
from .taskspawn import TaskSpawn
from .enums import WorkerType, SpawnStatus, WorkerState
from .net_classes import WorkerResources

from typing import TYPE_CHECKING, Optional, Tuple
if TYPE_CHECKING:
    from .scheduler import Scheduler


class SchedulerTaskProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, scheduler: "Scheduler", limit=2**16):
        self.__logger = logging.get_logger('scheduler')
        self.__timeout = 300.0
        self.__reader = asyncio.StreamReader(limit=limit)
        self.__scheduler = scheduler
        self.__saved_references = []
        super(SchedulerTaskProtocol, self).__init__(self.__reader, self.connection_cb)

    async def connection_cb(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        #
        #
        # commands

        async def comm_ping():  # if command == b'ping'
            # when worker pings scheduler - scheduler returns the state it thinks the worker is in
            addr = await read_string()
            wid = await self.__scheduler.worker_id_from_address(addr)
            if wid is None:
                state = WorkerState.UNKNOWN
            else:
                state = await self.__scheduler.get_worker_state(wid)
            writer.write(struct.pack('>I', state.value))

        async def comm_pulse():  # elif command == b'pulse':
            writer.write(b'\1')

        async def comm__pulse3way_():  # WARNING: this is for tests only!
            writer.write(b'\1')
            await writer.drain()
            await reader.readexactly(1)
            writer.write(b'\2')

        async def comm_done():  # elif command == b'done':
            tasksize = struct.unpack('>Q', await reader.readexactly(8))[0]
            task = await reader.readexactly(tasksize)
            task = await invocationjob.InvocationJob.deserialize_async(task)
            stdout = await read_string()
            stderr = await read_string()
            await self.__scheduler.task_done_reported(task, stdout, stderr)
            writer.write(b'\1')

        async def comm_dropped():  # elif command == b'dropped':
            tasksize = struct.unpack('>Q', await reader.readexactly(8))[0]
            task = await reader.readexactly(tasksize)
            task = await invocationjob.InvocationJob.deserialize_async(task)
            stdout = await read_string()
            stderr = await read_string()
            await self.__scheduler.task_cancel_reported(task, stdout, stderr)
            writer.write(b'\1')

        async def comm_hello():  # elif command == b'hello':
            # worker reports for duty
            addr = await read_string()
            workertype: WorkerType = WorkerType(struct.unpack('>I', await reader.readexactly(4))[0])
            reslength = struct.unpack('>Q', await reader.readexactly(8))[0]
            worker_hardware: WorkerResources = WorkerResources.deserialize(await reader.readexactly(reslength))
            await self.__scheduler.add_worker(addr, workertype, worker_hardware, assume_active=True)
            writer.write(struct.pack('>Q', self.__scheduler.db_uid()))

        async def comm_bye():  # elif command == b'bye':
            # worker reports he's quitting
            addr = await read_string()
            await self.__scheduler.worker_stopped(addr)
            writer.write(b'\1')

        #
        # commands used mostly by lifeblood_connection
        #
        # spawn a child task for task being processed
        async def comm_spawn():  # elif command == b'spawn':
            tasksize = struct.unpack('>Q', await reader.readexactly(8))[0]
            taskspawn: TaskSpawn = TaskSpawn.deserialize(await reader.readexactly(tasksize))
            ret: Tuple[SpawnStatus, Optional[int]] = await self.__scheduler.spawn_tasks(taskspawn)
            writer.write(struct.pack('>I?Q', ret[0].value, ret[1] is not None, 0 if ret[1] is None else ret[1]))

        async def comm_node_name_to_id():  # elif command == b'nodenametoid':
            nodename = await read_string()
            self.__logger.debug(f'got {nodename}')
            ids = await self.__scheduler.node_name_to_id(nodename)
            self.__logger.debug(f'sending {ids}')
            writer.write(struct.pack('>' + 'Q'*(1+len(ids)), len(ids), *ids))

        async def comm_update_task_attributes():  # elif command == b'tupdateattribs':  # note - this one is the same as in scheduler_ui_protocol...
            task_id, update_data_size, strcount = struct.unpack('>QQQ', await reader.readexactly(24))
            attribs_to_update = await asyncio.get_event_loop().run_in_executor(None, pickle.loads, await reader.readexactly(update_data_size))
            attribs_to_delete = set()
            for _ in range(strcount):
                attribs_to_delete.add(await read_string())
            await self.__scheduler.update_task_attributes(task_id, attribs_to_update, attribs_to_delete)
            writer.write(b'\1')

        async def comm_get_task_state():  # elif command == b'gettaskstate':
            task_id = struct.unpack('>Q', await reader.readexactly(8))[0]
            fields_dict = await self.__scheduler.get_task_fields(task_id)
            data = await asyncio.get_event_loop().run_in_executor(None, str.encode,
                                                                  await asyncio.get_event_loop().run_in_executor(None, json.dumps, fields_dict),
                                                                  'UTF-8')
            writer.write(struct.pack('>Q', len(data)))
            writer.write(data)

        async def comm_task_name_to_id():  # elif command == b'tasknametoid':
            taskname = await read_string()
            self.__logger.debug(f'got {taskname}')
            ids = await self.__scheduler.task_name_to_id(taskname)
            self.__logger.debug(f'sending {ids}')
            writer.write(struct.pack('>' + 'Q'*(1+len(ids)), len(ids), *ids))

        #
        commands = {'ping': comm_ping,
                    'pulse': comm_pulse,
                    '_pulse3way_': comm__pulse3way_,  # WARNING: this is for tests only!
                    'done': comm_done,
                    'dropped': comm_dropped,
                    'hello': comm_hello,
                    'bye': comm_bye,
                    'spawn': comm_spawn,
                    'nodenametoid': comm_node_name_to_id,
                    'tupdateattribs': comm_update_task_attributes,
                    'gettaskstate': comm_get_task_state,
                    'tasknametoid': comm_task_name_to_id,
                    }
        #
        #

        async def read_string() -> str:
            strlen = struct.unpack('>Q', await reader.readexactly(8))[0]
            if strlen == 0:
                return ''
            return (await reader.readexactly(strlen)).decode('UTF-8')

        async def write_string(s: str):
            b = s.encode('UTF-8')
            writer.write(struct.pack('>Q', len(b)))
            writer.write(b)

        try:
            # TODO: see same todo in worker_task_protocol
            prot = await asyncio.wait_for(reader.readexactly(4), self.__timeout)
            if prot != b'\0\0\0\0':
                raise NotImplementedError()

            while True:
                try:
                    command: str = await read_string()
                except asyncio.IncompleteReadError:  # no command sent, connection closed
                    self.__logger.debug('connection closed')
                    break
                self.__logger.debug(f'scheduler got command: {command}')

                if command in commands:
                    await commands[command]()

                #
                # if conn is closed - result will be b'', but in mostl likely totally impossible case it can be unfinished command.
                # so lets just catch all
                elif reader.at_eof():
                    self.__logger.debug('connection closed')
                    return
                else:
                    raise NotImplementedError()
                await writer.drain()

        except asyncio.exceptions.TimeoutError as e:
            self.__logger.warning(f'connection timeout happened ({self.__timeout}s).')
        except ConnectionResetError as e:
            self.__logger.exception('connection was reset. disconnected %s', e)
        except ConnectionError as e:
            self.__logger.exception('connection error. disconnected %s', e)
        except Exception as e:
            self.__logger.exception('unknown error. disconnected %s', e)
            raise
        finally:
            writer.close()
            await writer.wait_closed()
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())


class SchedulerTaskClient:
    def write_string(self, s: str):
        b = s.encode('UTF-8')
        self.__writer.write(struct.pack('>Q', len(b)))
        self.__writer.write(b)

    def __init__(self, ip: str, port: int):
        self.__logger = logging.get_logger('worker')
        self.__conn_task = asyncio.open_connection(ip, port)
        self.__reader = None  # type: Optional[asyncio.StreamReader]
        self.__writer = None  # type: Optional[asyncio.StreamWriter]

    async def __aenter__(self) -> "SchedulerTaskClient":
        await self._ensure_conn_open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        self.__writer.close()
        await self.__writer.wait_closed()

    async def _ensure_conn_open(self):
        if self.__reader is not None:
            return
        self.__reader, self.__writer = await self.__conn_task
        self.__writer.write(b'\0\0\0\0')

    async def report_task_done(self, task: invocationjob.InvocationJob, stdout_file: str, stderr_file: str):
        await self._ensure_conn_open()
        self.write_string('done')
        taskserialized = await task.serialize_async()
        self.__writer.write(struct.pack('>Q', len(taskserialized)))
        self.__writer.write(taskserialized)
        for std_file in (stdout_file, stderr_file):
            async with aiofiles.open(std_file, 'r') as f:
                self.write_string(await f.read())
        await self.__writer.drain()
        # we DO need a reply to ensure proper sequence of events
        assert await self.__reader.readexactly(1) == b'\1'

    async def report_task_canceled(self, task: invocationjob.InvocationJob, stdout_file: str, stderr_file: str):
        await self._ensure_conn_open()
        self.write_string('dropped')
        taskserialized = await task.serialize_async()
        self.__writer.write(struct.pack('>Q', len(taskserialized)))
        self.__writer.write(taskserialized)
        for std_file in (stdout_file, stderr_file):
            async with aiofiles.open(std_file, 'r') as f:
                self.write_string(await f.read())
        await self.__writer.drain()
        # we DO need a reply to ensure proper sequence of events
        assert await self.__reader.readexactly(1) == b'\1'

    async def ping(self, my_address: str) -> WorkerState:
        """
        remind scheduler about worker's existence and get back what he thinks of us

        :param my_address: address of this worker used to register at scheduler
        :return: worker state that scheduler thinks a worker with given address has
        """
        await self._ensure_conn_open()
        self.write_string('ping')
        try:
            self.write_string(my_address)
            await self.__writer.drain()
            return WorkerState(struct.unpack('>I', await self.__reader.readexactly(4))[0])
        except ConnectionResetError as e:
            self.__logger.error('ping failed. %s', e)
            raise

    async def pulse(self) -> None:
        """
        just ping the scheduler and get back a response, check if it's alive
        check pulse sorta

        :return:
        """
        await self._ensure_conn_open()
        self.write_string('pulse')
        try:
            await self.__writer.drain()
            await self.__reader.readexactly(1)
        except ConnectionResetError as e:
            self.__logger.error('pulse check failed. %s', e)
            raise

    async def _pulse3way_(self):
        """
        FOR TESTS ONLY
        """
        await self._ensure_conn_open()
        self.write_string('_pulse3way_')
        await self.__writer.drain()
        yield
        await self.__reader.readexactly(1)
        yield
        self.__writer.write(b'\1')
        await self.__writer.drain()
        await self.__reader.readexactly(1)

    async def say_hello(self, address_to_advertise: str, worker_type: WorkerType, worker_resources: WorkerResources):
        await self._ensure_conn_open()
        self.write_string('hello')
        self.write_string(address_to_advertise)
        self.__writer.write(struct.pack('>I', worker_type.value))
        resdata = worker_resources.serialize()
        self.__writer.write(struct.pack('>Q', len(resdata)))
        self.__writer.write(resdata)
        await self.__writer.drain()
        # as return we get the database's unique id to distinguish it from others
        return struct.unpack('>Q', await self.__reader.readexactly(8))[0]

    async def say_bye(self, address_of_worker: str):
        await self._ensure_conn_open()
        self.write_string('bye')
        self.write_string(address_of_worker)
        await self.__writer.drain()
        # we DO need a reply to ensure proper sequence of events
        assert await self.__reader.readexactly(1) == b'\1'

    async def spawn(self, taskspawn: TaskSpawn) -> Tuple[SpawnStatus, Optional[int]]:
        await self._ensure_conn_open()
        self.write_string('spawn')
        data_ser = await taskspawn.serialize_async()
        self.__writer.write(struct.pack('>Q', len(data_ser)))
        self.__writer.write(data_ser)
        await self.__writer.drain()
        status, is_null, val = struct.unpack('>I?Q', await self.__reader.readexactly(13))
        return SpawnStatus(status), None if is_null else val



from copy import copy
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


def node_class():
    return Mantra


class Mantra(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'mantra'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'houdini', 'mantra', 'ifd', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'mantra'

    def __init__(self, name):
        super(Mantra, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('ifd path', 'ifd file path', NodeParameterType.STRING, "`task['file']`")
            ui.add_parameter('image path', 'output image file path', NodeParameterType.STRING, "`task['outimage']`")
            ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        env = InvocationEnvironment()

        if context.param_value('skip if exists'):
            script = 'import os\n' \
                     'if not os.path.exists({imgpath}):\n' \
                     '    import sys\n' \
                     '    from subprocess import Popen\n' \
                     "    sys.exit(Popen(['mantra', '-V', '2a', {ifdpath}, {imgpath}]).wait())\n" \
                     "else:\n" \
                     "    print('image file already exists, skipping work')\n" \
                    .format(imgpath=repr(context.param_value('image path')),
                            ifdpath=repr(context.param_value('ifd path')))

            invoc = InvocationJob(['python', ':/mantracall.py'])
            invoc.set_extra_file('mantracall.py', script)
        else:
            invoc = InvocationJob(['mantra', '-V', '2a', context.param_value('ifd path'), context.param_value('image path')], env=env)
        res = ProcessingResult(invoc)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        res.set_attribute('file', context.param_value('image path'))
        return res



import asyncio
import json
import uuid
from ..client import MessageClient, MessageClientFactory
from ..queue import MessageQueue
from ..interfaces import MessageStreamFactory
from ..messages import Message, MessageType
from ..address import AddressChain

from typing import Optional


class JsonMessageWrapper(Message):
    def __init__(self, message: Message):
        super().__init__(b'', MessageType.DEFAULT_MESSAGE, AddressChain(''), AddressChain(''), None)
        self.copy_from(message)
        self.__wrapped_message = message
        self.__body_as_json = None

    def __getattr__(self, item):
        return getattr(self.__wrapped_message, item)

    async def message_body_as_json(self) -> dict:
        if self.__body_as_json is None:
            self.__body_as_json = await asyncio.get_event_loop().run_in_executor(None,
                                                                                 lambda s: json.loads(bytes(s).decode('utf-8')),
                                                                                 self.message_body())
        return self.__body_as_json

    async def set_message_body_as_json(self, body: dict):
        self.__body_as_json = body
        asyncio.get_event_loop().run_in_executor(None,
                                                 lambda d: self.set_message_body(json.dumps(d).encode('utf-8')),
                                                 body)


class JsonMessageClient(MessageClient):
    async def send_message_as_json(self, data: dict) -> JsonMessageWrapper:
        return JsonMessageWrapper(await self.send_message(json.dumps(data).encode('utf-8')))

    async def receive_message(self, timeout: Optional[float] = 90) -> JsonMessageWrapper:
        message = await super().receive_message(timeout)
        return JsonMessageWrapper(message)


class JsonMessageClientFactory(MessageClientFactory):
    def create_message_client(self, queue: MessageQueue, session: uuid.UUID, *,
                              source_address_chain: AddressChain,
                              destination_address_chain: AddressChain,
                              message_stream_factory: MessageStreamFactory,
                              send_retry_attempts: int = 2) -> MessageClient:
        return JsonMessageClient(queue, session,
                                 source_address_chain=source_address_chain,
                                 destination_address_chain=destination_address_chain,
                                 message_stream_factory=message_stream_factory,
                                 send_retry_attempts=send_retry_attempts
                                 )


class CommandJsonMessageClient(JsonMessageClient):
    async def send_command(self, command: str, arguments: dict):
        return await self.send_message_as_json({
            'command': {
                'name': command,
                'arguments': arguments
            }
        })


class CommandJsonMessageClientFactory(JsonMessageClientFactory):
    def create_message_client(self, queue: MessageQueue, session: uuid.UUID, *,
                              source_address_chain: AddressChain,
                              destination_address_chain: AddressChain,
                              message_stream_factory: MessageStreamFactory,
                              send_retry_attempts: int = 2) -> MessageClient:
        return CommandJsonMessageClient(queue, session,
                                        source_address_chain=source_address_chain,
                                        destination_address_chain=destination_address_chain,
                                        message_stream_factory=message_stream_factory,
                                        send_retry_attempts=send_retry_attempts
                                        )




import asyncio
import aiosqlite
from functools import partial
from typing import Optional, Dict, List, Callable
from .config import get_config
from .logging import get_logger
from dataclasses import dataclass


@dataclass
class ConnectionPoolEntry:
    con: aiosqlite.Connection
    __count: int = 1
    __count_max: int = 1
    __total_usages: int = 1
    closer_task: Optional[asyncio.Task] = None
    need_to_commit: bool = False
    __close_callbacks: Optional[List[Callable]] = None
    do_close: bool = False

    @property
    def count(self) -> int:
        return self.__count

    @count.setter
    def count(self, val: int) -> None:
        self.__count_max = max(self.__count_max, val)
        if val > self.__count:
            self.__total_usages += val - self.__count
        self.__count = val

    @property
    def count_max(self) -> int:
        return self.__count_max

    @property
    def total_usages(self) -> int:
        return self.__total_usages

    def add_close_callback(self, callback: Callable, *args, **kwargs) -> None:
        if self.__close_callbacks is None:
            self.__close_callbacks = []
        if len(args) or len(kwargs):
            callback = partial(callback, *args, **kwargs)
        self.__close_callbacks.append(callback)

    def add_close_callback_if_not_in(self, callback: Callable) -> None:
        if self.__close_callbacks is not None and callback in self.__close_callbacks:
            return
        self.add_close_callback(callback)

    def _invoke_close_callbacks(self):
        if self.__close_callbacks is None:
            return
        try:
            for cb in self.__close_callbacks:
                cb()
        except Exception:
            get_logger('ConnectionPoolEntry').exception('problem running close connection callbacks. callback chain aborted.')
        finally:
            self.__close_callbacks = None  # why? dunno


class ConnectionPool:
    default_period = get_config('scheduler').get_option_noasync('shared_connection.keep_open_period', 0.0125)
    get_logger('lazy_connection_pool').debug(f'using default open period: {default_period}')

    def __init__(self):
        self.connection_cache: Dict[tuple, ConnectionPoolEntry] = {}
        self.pool_lock = asyncio.Lock()
        self.keep_open_period = self.default_period
        self.__logger = get_logger('shared_aiosqlite_connection')
        self.__my_loop = asyncio.get_running_loop()

    def is_in_this_loop(self, loop=None):
        """
        check if given loop is the one this object was created in
        :param loop: loop or None - then current loop is checked
        :return:
        """
        return self.__my_loop == loop or asyncio.get_running_loop()

    async def connection_closer(self, key):
        await asyncio.sleep(self.keep_open_period)
        self.__logger.debug('shared connection lifetime reached, will close')
        async with self.pool_lock:
            entry = self.connection_cache[key]
            if entry.count > 0:
                entry.do_close = True
                self.__logger.debug('active connections present, will close after last one exits')
            else:  # only == 0 possible
                await self._closer_inner(key)

    async def _closer_inner(self, key):
        assert self.pool_lock.locked()  # cannot quite check if it's locked specifically for us, so this sanity check is partial at best
        entry = self.connection_cache[key]
        try:
            if entry.need_to_commit:
                await entry.con.commit()
        finally:
            await entry.con.close()
            del self.connection_cache[key]
            get_logger('shared_aiosqlite_connection').debug(f'closing shared connection used by total {entry.total_usages}, max {entry.count_max} at the same time. left cached conns: {len(self.connection_cache)}')
            entry._invoke_close_callbacks()


class SharedLazyAiosqliteConnection:
    """
    this will NOT work if used in different threads with the same pool!
    collects multiple commits in one.
    be VERY mindful where you use it!
    It was intended to collect the work of multiple similar independent tasks working in parallel
    to avoid hundreds of commits per second

    NOTE! To avoid confusion - row_factory of new connection is always set to ROW
    Be VERY mindful of what uses these shared connections!
    """
    connection_pools = {}

    def __init__(self, conn_pool, db_path: str, cache_name: str, *args, **kwargs):
        if conn_pool is not None:
            raise NotImplementedError()
        loop = asyncio.get_running_loop()
        if conn_pool is None and SharedLazyAiosqliteConnection.connection_pools.get(loop) is None:
            SharedLazyAiosqliteConnection.connection_pools[loop] = ConnectionPool()
            get_logger('shared_aiosqlite_connection').debug('default connection pool created')
        self.__pool = SharedLazyAiosqliteConnection.connection_pools[loop]

        self.__db_path = db_path
        self.__con_args = args
        self.__con_kwargs = kwargs
        self.__cache_key = (self.__db_path, cache_name, tuple(args), tuple(kwargs.items()))
        self.__con: Optional[aiosqlite.Connection] = None

    async def __aenter__(self):
        # TODO: CBB: currently constant stream of overlapping transactions can keep one connection open forever
        #  luckily currently we cannot have that cuz this lazy conn is only used by _awaiter, and it's transactions are
        #  additionally serialized by a lock.
        #  btw with lazy transactions like this we can remove that lock (the idea for it was that sqlite's locking is faaar slower than explicit threading lock, but if it's all one same transaction - no need in locking at all)
        #  and gain probably a lot of performance
        #  so
        #  the problem can be fixed if all instances remember entry at __init__ or __aenter__ and not use it in __aexit__ or anywhere else after
        #  then that closer task can delete connection_cache entry, keeping it internaly till everyone exits
        #  then new transaction with same key will not see existing cache and create a new entry.
        async with self.__pool.pool_lock:
            if self.__cache_key in self.__pool.connection_cache:
                self.__con = self.__pool.connection_cache[self.__cache_key].con
                self.__pool.connection_cache[self.__cache_key].count += 1
                return self
            self.__con = aiosqlite.connect(self.__db_path, *self.__con_args, **self.__con_kwargs)
            self.__pool.connection_cache[self.__cache_key] = ConnectionPoolEntry(con=self.__con)
            self.__pool.connection_cache[self.__cache_key].con = self.__con
            self.__pool.connection_cache[self.__cache_key].count = 1
            self.__pool.connection_cache[self.__cache_key].closer_task = asyncio.create_task(self.__pool.connection_closer(self.__cache_key))
            await self.__con
            await self.__con.execute('PRAGMA synchronous=NORMAL')  # TODO: take this from config
            self.__con.row_factory = aiosqlite.Row
            return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self.__pool.pool_lock:
            entry = self.__pool.connection_cache[self.__cache_key]
            entry.count -= 1
            assert entry.count >= 0
            if entry.count > 0:
                return
            # so here only count == 0
            if entry.do_close:  # so closer task has already finished, and we need to do it's job here
                await self.__pool._closer_inner(self.__cache_key)

    def add_after_commit_callback(self, callable: Callable, *args, **kwargs):
        entry = self.__pool.connection_cache.get(self.__cache_key)
        if entry is None:
            raise RuntimeError('cannot add callback to already closed and finalized lazy transactions')
        entry.add_close_callback(callable, *args, **kwargs)

    async def commit(self, callback=None):
        """
        optional callback argument for convenience.
        will be treated as unique callback, so same element will be added only once

        :param callback: callback to be launched right after db connection close
        :return:
        """
        entry = self.__pool.connection_cache[self.__cache_key]
        entry.need_to_commit = True
        if callback is not None:
            entry.add_close_callback_if_not_in(callback)

    async def rollback(self):
        raise RuntimeError('cannot rollback on shared connection')

    def __getattr__(self, item):
        return getattr(self.__con, item)



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult

from typing import Iterable


def node_class():
    return NullNode


class NullNode(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'null'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'null', 'passthrough', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'null'

    def __init__(self, name: str):
        super(NullNode, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.22, 0.24, 0.22)

    def process_task(self, context) -> ProcessingResult:
        return ProcessingResult()

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import asyncio
import json
import uuid
from . import broadcasting
from . import logging
from .nethelpers import get_localhost

from typing import Optional, Tuple, Callable, Coroutine, Any
from .logging import get_logger


class LocalMessageExchanger:
    __logger = None

    def __init__(self, identifier: str, callback: Callable[[Optional[str]], Coroutine]):
        self.__identifier = identifier
        self.__callback = callback
        self.__main_task = None
        self.__stop_event = asyncio.Event()
        self.__processed_signals_backlog_size = 16
        self.__processed_signals = []
        if self.__logger is None:
            LocalMessageExchanger.__logger = get_logger('LocalMessageExchanger')

    def start(self):
        if self.__main_task is not None:
            return
        self.__main_task = asyncio.create_task(self.__main_body())

    def stop(self):
        if self.__main_task is None:
            return
        self.__stop_event.set()

    async def wait_till_stopped(self):
        if self.__main_task is None:
            return
        await self.__main_task

    async def __main_body(self):
        """
        logic is this:
            - await for signal
            - initiate callback
         |->- get back on listening
         |  - if new signals appear during active callback - remember only the LATEST
         |  - if active callback is done and there is remembered signal - initiate callback for that too
         |--- and so on
        """
        magic_port = 11235
        addr = get_localhost()
        stopper = asyncio.create_task(self.__stop_event.wait())
        active_callback = None
        new_callback_needed: Tuple[bool, Optional[str]] = (False, None)  # with these two we will ensure no message gets lost DURING callback
        while not self.__stop_event.is_set():
            broadcaster_waiter = asyncio.create_task(broadcasting.await_broadcast(self.__identifier, magic_port, listen_address=addr))

            awaited = [stopper, broadcaster_waiter]
            if active_callback is not None:
                awaited.append(active_callback)
            done, _ = await asyncio.wait(awaited, return_when=asyncio.FIRST_COMPLETED)
            if stopper in done:
                broadcaster_waiter.cancel()
                if active_callback is not None:
                    new_callback_needed = (False, None)
                    await active_callback
                return
            if active_callback in done:
                active_callback = None
                if new_callback_needed[0]:
                    active_callback = asyncio.create_task(self.__callback(new_callback_needed[1]))
                    new_callback_needed = (False, None)
                if broadcaster_waiter not in done:
                    continue
            data = json.loads(await broadcaster_waiter)
            # create a new broadcast awaiter straight away

            self.__logger.debug(f'got {repr(data)}')
            await asyncio.sleep(0.01)  # we wait a tiny bit to prevent catching our own packet. it prevents a bit of flooding

            received_signal_id = data['n']
            message = data['m']
            if received_signal_id in self.__processed_signals:  # ignore
                continue
            self.__processed_signals.append(received_signal_id)
            if len(self.__processed_signals) > self.__processed_signals_backlog_size:
                self.__processed_signals = self.__processed_signals[-self.__processed_signals_backlog_size:]
            if active_callback is not None:
                new_callback_needed = (True, message)
            else:
                active_callback = asyncio.create_task(self.__callback(message))
                # TODO: we must ensure that callback is started only after broadcaster is listening again
                #  but here we do not ensure that

    async def send_sync_event(self, extra_message=None):
        await broadcasting.create_broadcaster(self.__identifier, json.dumps({'n': uuid.uuid4().hex,
                                                                             'm': extra_message}),
                                              broad_port=11235,
                                              ip=get_localhost(),
                                              broadcasts_count=1)


if __name__ == '__main__':  # a little test
    async def callback_test(message):
        print(f'callback! got {message}')

    async def main():
        mess = LocalMessageExchanger('test_exchanger', callback_test)
        mess.start()
        await asyncio.sleep(4)
        await mess.send_sync_event()
        await asyncio.sleep(4)
        mess.stop()
        print('waiting till stopped')
        await mess.wait_till_stopped()

    get_logger('LocalMessageExchanger').setLevel('DEBUG')
    asyncio.run(main())



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return BlenderScript


class BlenderScript(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'blender scene script'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'blender', 'script', 'python', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'blender_scene_script'

    @classmethod
    def description(cls) -> str:
        return 'This node runs given python script on a given blender scene.\n'

    def __init__(self, name):
        super(BlenderScript, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.922, 0.467, 0.0)
            ui.color_scheme().set_secondary_color(0.129, 0.337, 0.533)

            ui.add_parameter('file_path', 'File Path', NodeParameterType.STRING, "`task['file']`")
            ui.add_parameter('script', None, NodeParameterType.STRING, '# task var is available here\n# get task attribute with task[\'attr_name\']\n\n')\
                .set_text_multiline('python')

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        file_path = context.param_value('file_path')

        script_prefix = f'class _TaskWrapper:\n' \
                        f'    def __init__(self):\n' \
                        f'        self.__attrs = {repr(dict(context.task_attributes()))}\n' \
                        f'    @property\n' \
                        f'    def name(self):\n' \
                        f'        return {repr(context.task_field("name"))}\n' \
                        f'    @property\n' \
                        f'    def id(self):\n' \
                        f'        return {repr(context.task_id())}\n' \
                        f'    def attrib_names(self):\n' \
                        f'        return tuple(self.__attrs.keys())\n' \
                        f'    def __getitem__(self, key):\n' \
                        f'        return self.__attrs[key]\n' \
                        f'\n' \
                        f'task = _TaskWrapper()\n' \
                        f'\n\n'
        scripts = {}

        args = ['blender', '-b', file_path]

        scripts['script.py'] = script_prefix + context.param_value('script')
        args += ['-P', ':/script.py']

        invoc = InvocationJob(args)
        for file_name, file_contents in scripts.items():
            invoc.set_extra_file(file_name, file_contents)
        return ProcessingResult(invoc)



import asyncio
import aiofiles
import struct
import json
from .enums import TaskScheduleStatus, TaskExecutionStatus, TaskExecutionStatus, WorkerPingReply, SpawnStatus
from .exceptions import NotEnoughResources, ProcessInitializationError, WorkerNotAvailable, AlreadyRunning, CouldNotNegotiateProtocolVersion
from .scheduler_message_processor import SchedulerExtraControlClient, SchedulerInvocationMessageClient
from .environment_resolver import ResolutionImpossibleError
from .taskspawn import TaskSpawn
from . import logging
from . import invocationjob
from . import nethelpers

import os

from typing import Dict, Optional, Set, Sequence, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from .worker import Worker


async def read_string(reader) -> str:
    strlen = struct.unpack('>Q', await reader.readexactly(8))[0]
    if strlen == 0:
        return ''
    return (await reader.readexactly(strlen)).decode('UTF-8')


async def write_string(writer, s: str):
    b = s.encode('UTF-8')
    writer.write(struct.pack('>Q', len(b)))
    writer.write(b)
    

class ProtocolHandler:
    def protocol_version(self) -> Tuple[int, int]:
        raise NotImplementedError()

    async def handle(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        raise NotImplementedError()


class WorkerInvocationProtocolHandlerV10(ProtocolHandler):
    def __init__(self, worker: "Worker"):
        super().__init__()
        self.__worker = worker
        self.__logger = logging.get_logger(f'worker.invoc_protocol_v{".".join(str(i) for i in self.protocol_version())}')

        self.__known_commands = {
            'spawn': self.comm_spawn,
            'tupdateattribs': self.comm_update_attributes,
            'sendinvmessage': self.comm_send_invocation_message,
            'recvinvmessage': self.comm_receive_invocation_message,
        }
        
    def protocol_version(self) -> Tuple[int, int]:
        return 1, 0

    async def handle(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        # TODO: either we need timeout here - then add it to all reading operations below
        # TODO: or we don't need it here - then remove wait.
        command = await read_string(reader)
        self.__logger.debug(f'got command: {command}')

        if command not in self.__known_commands:
            raise NotImplementedError(f'unknown command {command}')
        await self.__known_commands[command](reader, writer)

    #
    # commands
    async def comm_spawn(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        tasksize = struct.unpack('>Q', await reader.readexactly(8))[0]
        taskspawn: TaskSpawn = TaskSpawn.deserialize(await reader.readexactly(tasksize))

        with SchedulerExtraControlClient.get_scheduler_control_client(self.__worker.scheduler_message_address(),
                                                                      self.__worker.message_processor()) as client:  # type: SchedulerExtraControlClient
            status, tid = await client.spawn(taskspawn)
        writer.write(struct.pack('>I?Q', status.value, tid is not None, 0 if tid is None else tid))

    async def comm_update_attributes(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        task_id, update_data_size, strcount = struct.unpack('>QQQ', await reader.readexactly(24))
        attribs_to_update = await asyncio.get_event_loop().run_in_executor(None, json.loads,
                                                                           (await reader.readexactly(update_data_size)).decode('UTF-8'))
        attribs_to_delete = set()
        for _ in range(strcount):
            attribs_to_delete.add(await read_string(reader))
        with SchedulerExtraControlClient.get_scheduler_control_client(self.__worker.scheduler_message_address(),
                                                                      self.__worker.message_processor()) as client:  # type: SchedulerExtraControlClient
            await client.update_task_attributes(task_id, attribs_to_update, attribs_to_delete)
        writer.write(b'\1')

    # invocation messaging
    async def comm_send_invocation_message(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        to_inv_id, from_inv_id, data_size, addressee_timeout = struct.unpack('>QQQf', await reader.readexactly(28))
        to_addressee = await read_string(reader)
        data = await reader.readexactly(data_size)
        with SchedulerInvocationMessageClient.get_scheduler_control_client(self.__worker.scheduler_message_address(),
                                                                           self.__worker.message_processor()) as client:  # type: SchedulerInvocationMessageClient
            send_status = await client.send_invocation_message(to_inv_id, to_addressee, from_inv_id, data, addressee_timeout=addressee_timeout)
        await write_string(writer, send_status.value)

    async def comm_receive_invocation_message(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        addressee = await read_string(reader)
        poll_interval, = struct.unpack('>f', await reader.readexactly(4))
        data = None
        src_iid = None
        while data is None:
            try:
                src_iid, data = await self.__worker.worker_task_addressee_wait(addressee, timeout=poll_interval)
            except asyncio.TimeoutError:
                # in case of timeout - we keep waiting, but better poke that receiver still lives
                writer.write(struct.pack('>?', False))
                await writer.drain()
                cancelled, = struct.unpack('>?', await reader.readexactly(1))
                if cancelled:
                    return
        assert src_iid is not None
        writer.write(struct.pack('>?QQ', True, src_iid, len(data)))
        writer.write(data)
        await writer.drain()


class WorkerInvocationServerProtocol(asyncio.StreamReaderProtocol):

    def __init__(self, worker: "Worker", protocol_handlers: Sequence[ProtocolHandler], limit: int = 2 ** 16):
        self.__logger = logging.get_logger('worker.invoc_protocol')
        self.__timeout = 300.0
        self.__reader = asyncio.StreamReader(limit=limit)
        self.__worker = worker
        self.__handlers: Dict[Tuple[int, int], ProtocolHandler] = {x.protocol_version(): x for x in protocol_handlers}
        self.__supported_versions = tuple(sorted(self.__handlers.keys(), reverse=True))
        self.__saved_references = []
        super(WorkerInvocationServerProtocol, self).__init__(self.__reader, self.client_connected_cb)

    def supported_protocol_versions(self) -> Tuple[Tuple[int, int], ...]:
        """
        return supported versions, sorted from highest to lowest
        """
        return self.__supported_versions

    async def __negotiate_protocol_version(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> Tuple[int, int]:
        num_protocols, = struct.unpack('>Q', await reader.readexactly(8))
        vers: Set[Tuple[int, int]] = set()
        for _ in range(num_protocols):
            vmaj, vmin = struct.unpack('>II', await reader.readexactly(8))
            assert isinstance(vmaj, int) and isinstance(vmin, int)
            vers.add((vmaj, vmin))

        for ver in self.supported_protocol_versions():
            if ver in vers:
                writer.write(struct.pack('>II', *ver))
                await writer.drain()
                return ver
        raise CouldNotNegotiateProtocolVersion(self.supported_protocol_versions(), tuple(vers))

    async def client_connected_cb(self, reader, writer):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        try:
            await self.__worker.wait_till_starts()  # important that worker is fully up before we actually start listening to shit

            prot = await self.__negotiate_protocol_version(reader, writer)
            await self.__handlers[prot].handle(reader, writer)

            # drain just in case
            await writer.drain()
        except asyncio.TimeoutError:
            self.__logger.error('operation timeout')
            raise
        except EOFError:
            self.__logger.error('connection was abruptly closed')
            raise
        finally:
            writer.close()
            await writer.wait_closed()
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())

#
# class WorkerInvocationClient:
#     def __init__(self, ip: str, port: int, timeout=300.0):
#         self.__logger = logging.get_logger('scheduler.invoc_connection')
#         self.__conn_task = asyncio.create_task(asyncio.open_connection(ip, port))
#         self.__reader = None  # type: Optional[asyncio.StreamReader]
#         self.__writer = None  # type: Optional[asyncio.StreamWriter]
#         self.__timeout = timeout
#
#     async def __aenter__(self) -> "WorkerInvocationClient":
#         await self._ensure_conn_open()
#         return self
#
#     async def __aexit__(self, exc_type, exc_val, exc_tb):
#         await self.close()
#
#     async def close(self):
#         self.__writer.close()
#         await self.__writer.wait_closed()
#
#     async def _ensure_conn_open(self):
#         if self.__reader is not None:
#             return
#         self.__reader, self.__writer = await self.__conn_task
#         self.__writer.write(b'')
#
#     raise NotImplementedError()



from enum import Enum


class MessageType(Enum):
    @classmethod
    def message_type_size(cls) -> int:
        return 2

    DEFAULT_MESSAGE = b'\0\0'
    SYSTEM_PING = b'\0\1'
    DELIVERY_ERROR = b'\1\0'



import uuid
from .exceptions import MessageSendingError, MessageReceivingError, MessageRoutingError, MessageTransferError
from .messages import Message
from .address import AddressChain, DirectAddress
from .enums import MessageType
from .defaults import default_stream_timeout

from typing import Optional


class MessageSendStreamBase:
    """
    Represents a direct channel between source and destination
    """
    def __init__(self, reply_address: DirectAddress, destination_address: DirectAddress, 
                 stream_timeout: Optional[float] = default_stream_timeout,
                 message_confirmation_timeout: Optional[float] = default_stream_timeout):
        self.__source = reply_address
        self.__destination = destination_address
        self.__timeout = stream_timeout if stream_timeout and stream_timeout > 0 else None
        self.__confirmation_timeout = message_confirmation_timeout
        self.__ping_timeout = message_confirmation_timeout

    def _stream_timeout(self) -> Optional[float]:
        return self.__timeout

    def _confirmation_timeout(self) -> Optional[float]:
        """
        how long to wait from message sending till confirmation receiving
        """
        return self.__confirmation_timeout

    def reply_address(self) -> DirectAddress:
        return self.__source

    def destination_address(self) -> DirectAddress:
        return self.__destination

    async def _send_message_implementation(self, message: Message, timeout: Optional[float]):
        """
        override this with actual message sending implementation

        on timeout should raise MessageTransferTimeoutError
        """
        raise NotImplementedError()

    async def _ack_message_implementation(self, message: Message, timeout: Optional[float]):
        """
        implementation-specific ack received after message is sent

        on timeout should raise MessageTransferTimeoutError
        """
        raise NotImplementedError()

    async def send_raw_message(self, message: Message, *, message_delivery_timeout_override: Optional[float] = ...):
        """
        send message as is

        :raises MessageSendingError: wrapping actual error occured
        """
        if message.message_destination().split_address()[0] != self.destination_address():
            raise MessageRoutingError('message destination address must contain stream\'s destination')
        if message.message_source().split_address()[0] != self.reply_address():
            raise MessageRoutingError('message source address must contain stream\'s source')
        try:
            await self._send_message_implementation(message, self._stream_timeout())
            await self._ack_message_implementation(message, self._confirmation_timeout() if message_delivery_timeout_override is ... else message_delivery_timeout_override)
        except MessageTransferError:
            raise
        except Exception as e:
            raise MessageSendingError(wrapped_exception=e) from None

    async def send_ping(self):
        await self.send_raw_message(Message(b'', MessageType.SYSTEM_PING, self.__source, self.__destination, None),
                                    message_delivery_timeout_override=self.__ping_timeout)

    async def send_data_message(self, data: bytes, destination: AddressChain, *, session: uuid.UUID, source: Optional[AddressChain] = None) -> Message:
        """
        send message through stream

        :raises MessageSendingError: wrapping actual error occured
        :raises MessageRoutingError: if destination does not contain this stream's destination
        """
        source = source or self.reply_address()
        message = Message(data, MessageType.DEFAULT_MESSAGE, source, destination, session)
        await self.send_raw_message(message)
        return message

    def close(self):
        """
        start stream closing process
        if stream is already closing - do nothing
        """
        raise NotImplementedError()

    async def wait_closed(self):
        """
        wait till stream is closed
        if stream is already closed - return immediately
        """
        raise NotImplementedError()


class MessageReceiveStreamBase:
    def __init__(self, listening_address: DirectAddress, other_end_address: DirectAddress, stream_timeout: Optional[float] = default_stream_timeout):
        self.__source = listening_address
        self.__destination = other_end_address
        self.__message_acked = True
        self.__timeout = stream_timeout if stream_timeout and stream_timeout > 0 else None

    def _stream_timeout(self) -> Optional[float]:
        return self.__timeout

    def listening_address(self) -> DirectAddress:
        return self.__source

    def other_end_address(self) -> DirectAddress:
        return self.__destination

    async def _receive_message_implementation(self) -> Message:
        """
        override this with actual message receiving implementation

        other implementation-specific exceptions will be wrapped into MessageReceivingError by receive_data_message

        :raises NoMessage: special case of timeout error when no data was read
        :raises MessageTransferTimeoutError: timeout error, when partial data was read
        """
        raise NotImplementedError()

    async def _acknowledge_message_implementation(self, status: bool):
        """
        override this with actual message acknowledgement implementation

        implementation-specific exceptions will be wrapped into MessageReceivingError by acknowledge_received_message
        """
        raise NotImplementedError()

    async def receive_data_message(self) -> Message:
        """
        receive message from stream

        :raises MessageReceivingError: wrapping actual error occured
        :returns: Message received from the stream
        """
        if not self.__message_acked:
            raise RuntimeError('previous message was not acknowledged')
        try:
            message = await self._receive_message_implementation()
        except MessageTransferError:
            raise
        except Exception as e:
            raise MessageReceivingError(wrapped_exception=e) from None
        else:
            self.__message_acked = False
            return message

    async def acknowledge_received_message(self, success: bool):
        if self.__message_acked:
            raise RuntimeError('there was no message to acknowledge')
        try:
            await self._acknowledge_message_implementation(success)
        except MessageTransferError:
            raise
        except Exception as e:
            raise MessageTransferError(wrapped_exception=e) from None
        self.__message_acked = True

    def close(self):
        """
        start stream closing process
        if stream is already closing - do nothing
        """
        raise NotImplementedError()

    async def wait_closed(self):
        """
        wait till stream is closed
        if stream is already closed - return immediately
        """
        raise NotImplementedError()



"""
this module contains TaskSpawn class used from scheduler nodes and
worker jobs to spawn new tasks
"""
import pickle
import asyncio
from io import BytesIO
from .environment_resolver import EnvironmentResolverArguments

from typing import Optional, Tuple, List, Iterable


class Unpickler(pickle.Unpickler):
    def find_class(self, module, name):
        module_parts = module.split('.')
        if module_parts[-1] == 'lifeblood_connection' or \
           module_parts[-2:] == ['lifeblood_client', 'submitting']:  # TODO: this becomes dirty as it scales... make this more generic!
            if name == 'TaskSpawn':
                return TaskSpawn
            elif name == 'NewTask':
                return NewTask
            elif name == 'EnvironmentResolverArguments':
                return EnvironmentResolverArguments
        return super(Unpickler, self).find_class(module, name)


class TaskSpawn:
    def __init__(self, name: str, source_invocation_id: Optional[int] = None, env_args: Optional[EnvironmentResolverArguments] = None, task_attributes: dict = None, extra_groups: Iterable[str] = None):
        """

        :param name:
        :param source_invocation_id:
        :param env_args: if None - will inherit parent's. if no parent and none, or if empty arguments- default worker's env wrapper will be used
        :param task_attributes:
        """
        self.__name = name
        self.__attributes = dict(task_attributes or {})
        self.__env_args = env_args
        self.__forced_node_task_id_pair = None
        self.__from_invocation_id = source_invocation_id
        self.__output = 'spawned'
        self._create_as_spawned = True
        self.__extra_groups = list(extra_groups) if extra_groups is not None else []
        self.__default_priority = None

    def create_as_spawned(self):
        return self._create_as_spawned

    def force_set_node_task_id(self, node_id: Optional[int], task_id: Optional[int]):
        """
        force override some standard inheritance

        :param node_id: if given - will be the node to spawn task in
        :param task_id: if given - will override the parent task for this spawning task (must be given together with node_id)
        :return:
        """
        if node_id is None and task_id is None:
            self.__forced_node_task_id_pair = None
        self.__forced_node_task_id_pair = (node_id, task_id)

    def forced_node_task_id(self) -> Optional[Tuple[int, int]]:
        return self.__forced_node_task_id_pair

    def source_invocation_id(self):
        return self.__from_invocation_id

    def set_node_output_name(self, new_name):
        self.__output = new_name

    def node_output_name(self) -> str:
        return self.__output

    def name(self) -> str:
        return self.__name

    def default_priority(self) -> Optional[float]:
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is

        :return: default priority
        """
        return self.__default_priority

    def add_extra_group_name(self, group_name: str) -> None:
        self.__extra_groups.append(group_name)

    def add_extra_group_names(self, group_names: Iterable[str]) -> None:
        self.__extra_groups += group_names

    def extra_group_names(self) -> List[str]:
        return self.__extra_groups

    def set_default_priority(self, priority: float):
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is
        """
        self.__default_priority = priority

    def set_name(self, name):
        self.__name = name

    def set_attribute(self, attr_name, attr_value):
        self.__attributes[attr_name] = attr_value

    def remove_attribute(self, attr_name):
        del self.__attributes[attr_name]

    def attribute_value(self, attr_name):
        return self.__attributes.get(attr_name, None)

    def _attributes(self):
        return self.__attributes

    def environment_arguments(self):
        return self.__env_args

    def set_environment_resolver(self, resolver_name, resolver_arguments):
        self.__env_args = EnvironmentResolverArguments(resolver_name, resolver_arguments)

    def serialize(self) -> bytes:
        return pickle.dumps(self)

    async def serialize_async(self) -> bytes:
        return await asyncio.get_event_loop().run_in_executor(None, self.serialize)

    @classmethod
    def deserialize(cls, data: bytes) -> "TaskSpawn":
        deserializer = Unpickler(BytesIO(data))
        return deserializer.load()

    @classmethod
    async def deserialize_async(cls, data: bytes) -> "TaskSpawn":
        return await asyncio.get_event_loop().run_in_executor(None, cls.deserialize, data)


class NewTask(TaskSpawn):
    def __init__(self, name: str, node_id: int, env_args: Optional[EnvironmentResolverArguments] = None, task_attributes: Optional[dict] = None, priority: float = 50.0, extra_groups: Iterable[str] = None):
        super(NewTask, self).__init__(name, None, env_args, task_attributes, extra_groups)
        self.set_node_output_name('main')
        self.force_set_node_task_id(node_id, None)
        self._create_as_spawned = False
        self.set_default_priority(priority)



import logging
from .ui_events import SchedulerEvent
from .enums import UIEventType

from typing import Dict, List, Optional, Set, Tuple


logger = logging.getLogger(__name__)


class SchedulerEventLog:
    def __init__(self, *, log_time_length_max: Optional[float] = None, log_event_count_max: Optional[int] = None):
        self.__max_time_ns = int(log_time_length_max * 10**9) if log_time_length_max is not None else None
        self.__max_count = log_event_count_max
        if log_event_count_max is None and log_time_length_max is None:
            raise ValueError('at least one of log_time_length_max or log_event_count_max must be not none')

        self.__events: List[SchedulerEvent] = []  # will be kept sorted
        self.__event_id_to_event: Dict[int, SchedulerEvent] = {}
        self.__full_state_events: List[SchedulerEvent] = []

        self.__internal_event_counter_next = 0

    def __find_index_below_timestamp(self, timestamp):
        return self.__find_index_below(self.__events, 'timestamp', timestamp)

    def __find_index_below_id(self, event_id):
        return self.__find_index_below(self.__events, 'event_id', event_id)

    def __find_full_state_index_below_id(self, event_id):
        return self.__find_index_below(self.__full_state_events, 'event_id', event_id)

    @classmethod
    def __find_index_below(cls, event_list, property_name, value):
        if len(event_list) < 8:
            for i, event in enumerate(event_list):
                if getattr(event, property_name) > value:
                    return i - 1
            return len(event_list) - 1

        def _find_bin(l: List[SchedulerEvent], value):
            if len(l) == 1:
                return -1 if value < getattr(l[0], property_name) else 0
            if value < getattr(l[0], property_name):
                return -1
            elif value > getattr(l[-1], property_name):
                return len(l)-1
            si = len(l) // 2
            if (i := _find_bin(l[si:], value)) > -1:
                return i + si
            return _find_bin(l[:si], value)

        return _find_bin(event_list, value)

    def add_event(self, event: SchedulerEvent, do_trim=True):
        # first check the trivial case
        if event.event_id < 0:
            event.event_id = self.__internal_event_counter_next
        if len(self.__events) == 0 or event.event_id > self.__events[-1].event_id and event.timestamp >= self.__events[-1].timestamp:
            i = len(self.__events)
        else:
            if event.event_id in self.__event_id_to_event:
                if event != self.__event_id_to_event[event.event_id]:
                    raise RuntimeError('attempting to add event with the same id, but different data')
                logger.warning(f'ignoring event as it is already in the log: {event}')
                return
            i = self.__find_index_below_id(event.event_id) + 1
            if i > 0 and (self.__events[i-1].event_id >= event.event_id or self.__events[i-1].timestamp > event.timestamp) \
                    or i < len(self.__events) and (self.__events[i].event_id <= event.event_id or self.__events[i].timestamp < event.timestamp):  # sanity checks
                raise RuntimeError(f'event IDs are expected to be sorted the same way as their timestamps: [{f"{self.__events[i-1].event_id}:{self.__events[i-1].timestamp}" if i>0 else "|"}, <{event.event_id}:{event.timestamp}>, {f"{self.__events[i].event_id}:{self.__events[i].timestamp}" if i<len(self.__events) else "|"}]')
        self.__events.insert(i, event)
        self.__event_id_to_event[event.event_id] = event
        if event.event_type == UIEventType.FULL_STATE:
            if len(self.__full_state_events) == 0 or i == len(self.__events) - 1:
                self.__full_state_events.append(event)
            else:
                self.__full_state_events.insert(self.__find_full_state_index_below_id(event.event_id) + 1, event)

        self.__internal_event_counter_next = max(self.__internal_event_counter_next, event.event_id) + 1
        if do_trim:
            self.trim()

    def trim(self):
        if len(self.__events) == 0:
            return
        if self.__max_time_ns is not None:
            threshold = self.__events[-1].timestamp - self.__max_time_ns
            if self.__events[0].timestamp <= threshold:
                trimpos = self.__find_index_below_timestamp(threshold)+1
                for event in self.__events[:trimpos]:
                    deleted_event = self.__event_id_to_event.pop(event.event_id)
                    if deleted_event.event_type == UIEventType.FULL_STATE:
                        # since we delete always last - we are guaranteed to have full events in the same order
                        assert deleted_event.event_id == self.__full_state_events.pop(0).event_id
                self.__events = self.__events[trimpos:]

        if self.__max_count is not None and len(self.__events) > self.__max_count:
            for event in self.__events[:-self.__max_count]:
                deleted_event = self.__event_id_to_event.pop(event.event_id)
                if deleted_event.event_type == UIEventType.FULL_STATE:
                    # since we delete always last - we are guaranteed to have full events in the same order
                    assert deleted_event.event_id == self.__full_state_events.pop(0).event_id
            self.__events = self.__events[-self.__max_count:]

    def get_since_timestamp(self, timestamp: float):
        """
        get events since given timestamp
        :param timestamp: timestamp is given as EPOCH time in floating seconds
        :return:
        """
        timestamp_ns = int(timestamp) * 10**9 + int(timestamp % 1.0 * 10**9)
        return self.get_since_timestamp_ns(timestamp_ns)

    def get_since_timestamp_ns(self, timestamp_ns: int):
        return tuple(self.__events[self.__find_index_below_timestamp(timestamp_ns)+1:])

    def get_since_event(self, event_id: int, truncate_before_full_state=False) -> Tuple[SchedulerEvent, ...]:

        # if event was already removed from the list (assume it was old)
        if truncate_before_full_state and len(self.__full_state_events) > 0:
            possible_full_state_i = self.__find_full_state_index_below_id(event_id) + 1
            if possible_full_state_i < len(self.__full_state_events):
                event_id = self.__full_state_events[possible_full_state_i].event_id - 1
        e_index = self.__find_index_below_id(event_id)
        if 0 <= e_index < len(self.__events):
            assert self.__events[e_index].event_id <= event_id  # in case event_id does not exist  - found event_id will be smaller
        # we will return all events since e_index, not since it's timestamp cuz we relay in incrementing event_ids
        # and want to avoid extra search by timestamp, since the result should be the exact same
        return tuple(self.__events[e_index+1:])

    def __len__(self):
        return len(self.__events)



import aiosqlite
import sqlite3
import random
import struct
from ..db_misc import sql_init_script
from ..logging import get_logger
from ..scheduler_event_log import SchedulerEventLog
from ..shared_lazy_sqlite_connection import SharedLazyAiosqliteConnection
from .. import aiosqlite_overlay

SCHEDULER_DB_FORMAT_VERSION = 1


class DataAccess:
    def __init__(self, db_path, db_connection_timeout):
        self.__logger = get_logger('scheduler.data_access')
        self.db_path: str = db_path
        self.db_timeout: int = db_connection_timeout

        self.mem_cache_workers_resources: dict = {}
        self.mem_cache_workers_state: dict = {}
        self.mem_cache_invocations: dict = {}
        #
        # ensure database is initialized
        with sqlite3.connect(db_path) as con:
            con.executescript(sql_init_script)
        with sqlite3.connect(db_path) as con:
            con.row_factory = sqlite3.Row
            cur = con.execute('SELECT * FROM lifeblood_metadata')
            metadata = cur.fetchone()  # there should be exactly one single row.
            cur.close()
            if metadata is None:  # if there's no - the DB has not been initialized yet
                # we need 64bit signed id to save into db
                db_uid = random.getrandbits(64)  # this is actual db_uid
                db_uid_signed = struct.unpack('>q', struct.pack('>Q', db_uid))[0]  # this one goes to db
                con.execute('INSERT INTO lifeblood_metadata ("version", "component", "unique_db_id")'
                            'VALUES (?, ?, ?)', (SCHEDULER_DB_FORMAT_VERSION, 'scheduler', db_uid_signed))
                con.commit()
                # reget metadata
                cur = con.execute('SELECT * FROM lifeblood_metadata')
                metadata = cur.fetchone()  # there should be exactly one single row.
                cur.close()
            self.__db_uid = struct.unpack('>Q', struct.pack('>q', metadata['unique_db_id']))[0]  # reinterpret signed as unsigned

    @property
    def db_uid(self):
        return self.__db_uid

    def data_connection(self) -> aiosqlite_overlay.ConnectionWithCallbacks:
        return aiosqlite_overlay.connect(self.db_path, timeout=self.db_timeout, pragmas_after_connect=('synchronous=NORMAL',))

    def lazy_data_transaction(self, key_name: str):
        return SharedLazyAiosqliteConnection(None, self.db_path, key_name, timeout=self.db_timeout)

    async def write_back_cache(self):
        self.__logger.info('pinger syncing temporary tables back...')
        async with self.data_connection() as con:
            for wid, cached_row in self.mem_cache_workers_state.items():
                await con.execute('UPDATE workers SET '
                                  'last_seen=?, '
                                  'last_checked=?, '
                                  'ping_state=? '
                                  'WHERE "id"=?',
                                  (cached_row['last_seen'], cached_row['last_checked'], cached_row['ping_state'], wid))
            await con.commit()



import struct
import socket


def recv_exactly(sock, numbytes):  # type: (socket.socket, int) -> bytes
    patches = []
    got_bytes = 0
    while got_bytes != numbytes:
        patches.append(sock.recv(numbytes - got_bytes))
        got_bytes += len(patches[-1])
        if len(patches[-1]) == 0:
            raise ConnectionResetError()

    if len(patches) == 1:
        return patches[0]
    elif len(patches) == 0:
        return b''
    return b''.join(patches)


def send_string(sock, text):  # type: (socket.socket, str) -> None
    bts = text.encode('UTF-8')
    sock.sendall(struct.pack('>Q', len(bts)))
    sock.sendall(bts)


def recv_string(sock):  # type: (socket.socket) -> str
    btlen = struct.unpack('>Q', recv_exactly(sock, 8))[0]
    return recv_exactly(sock, btlen).decode('UTF-8')



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.taskspawn import TaskSpawn
from lifeblood.exceptions import NodeNotReadyToProcess
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi, MultiGroupLayout, Parameter
from lifeblood.node_visualization_classes import NodeColorScheme

from typing import Iterable


def node_class():
    return ModifyAttributes


class ModifyAttributes(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'modify attributes'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'modify', 'attribute', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'mod_attrib'

    def __init__(self, name: str):
        super(ModifyAttributes, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.15, 0.24, 0.25)
            with ui.multigroup_parameter_block('attr_count'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('name', 'Name', NodeParameterType.STRING, '')
                    optype_param = ui.add_parameter('optype', '', NodeParameterType.INT, 0)
                    optype_param.add_menu((('add', 0),
                                           ('sub', 1),
                                           ('mul', 2),
                                           ('div', 3),
                                           ('concat', 100),
                                           ('append', 200),
                                           ('pop', 201),
                                           ('slice', 202),
                                           ))

                    valtype_param = ui.add_parameter('valtype', '', NodeParameterType.INT, NodeParameterType.INT.value)
                    valtype_param.append_visibility_condition(optype_param, '<', 100)
                    valtype_param.add_menu((('int', NodeParameterType.INT.value),
                                            ('float', NodeParameterType.FLOAT.value)))

                    ui.add_parameter('ivalue', 'val', NodeParameterType.INT, 0)\
                        .append_visibility_condition(optype_param, '<', 100)\
                        .append_visibility_condition(valtype_param, '==', NodeParameterType.INT.value)
                    ui.add_parameter('fvalue', 'val', NodeParameterType.FLOAT, 0.0)\
                        .append_visibility_condition(optype_param, '<', 100)\
                        .append_visibility_condition(valtype_param, '==', NodeParameterType.FLOAT.value)

                    ui.add_parameter('svalue', 'val', NodeParameterType.STRING, '')\
                        .append_visibility_condition(optype_param, '==', 100)

                    listvaltype_param = ui.add_parameter('lvaltype', '', NodeParameterType.INT, NodeParameterType.INT.value)\
                        .add_menu((('int', NodeParameterType.INT.value),
                                   ('float', NodeParameterType.FLOAT.value),
                                   ('string', NodeParameterType.STRING.value)))\
                        .append_visibility_condition(optype_param, '==', 200)
                    ui.add_parameter('lfvalue', 'val', NodeParameterType.FLOAT, 0.0) \
                        .append_visibility_condition(optype_param, '==', 200)\
                        .append_visibility_condition(listvaltype_param, '==', NodeParameterType.FLOAT.value)
                    ui.add_parameter('livalue', 'val', NodeParameterType.INT, 0) \
                        .append_visibility_condition(optype_param, '==', 200)\
                        .append_visibility_condition(listvaltype_param, '==', NodeParameterType.INT.value)
                    ui.add_parameter('lsvalue', 'val', NodeParameterType.STRING, '') \
                        .append_visibility_condition(optype_param, '==', 200)\
                        .append_visibility_condition(listvaltype_param, '==', NodeParameterType.STRING.value)

                    # slice
                    ui.add_parameter('lsl0value', '', NodeParameterType.INT, 0) \
                        .append_visibility_condition(optype_param, '==', 202)
                    ui.add_parameter('lsl1value', '', NodeParameterType.INT, 0) \
                        .append_visibility_condition(optype_param, '==', 202)
                    ui.add_parameter('lsl2value', 'start/end/step', NodeParameterType.INT, 1) \
                        .append_visibility_condition(optype_param, '==', 202)

    def process_task(self, context) -> ProcessingResult:
        attr_count = context.param_value('attr_count')
        res = ProcessingResult()
        to_set = {}
        for i in range(attr_count):
            attrs = context.task_attributes()
            attr_name = context.param_value(f'name_{i}')
            if attr_name == '':  # however some spaces - is a valid parameter name
                continue
            if attr_name in attrs:
                to_set[attr_name] = attrs[attr_name]

            op_type = context.param_value(f'optype_{i}')
            if op_type < 100:
                attr_type = context.param_value(f'valtype_{i}')
                if attr_type == NodeParameterType.INT.value:
                    attr_val = context.param_value(f'ivalue_{i}')
                    if attr_name not in to_set:
                        to_set[attr_name] = 0
                    # remember - sets should be independent, cuz if u do them dependant on the order -
                    #  the expression eval will still be independent and it's gonna be a confusing mess

                elif attr_type == NodeParameterType.FLOAT.value:
                    attr_val = context.param_value(f'fvalue_{i}')
                    if attr_name not in to_set:
                        to_set[attr_name] = 0.0
                else:
                    raise NotImplementedError()
                if op_type == 0:
                    to_set[attr_name] += attr_val
                elif op_type == 1:
                    to_set[attr_name] -= attr_val
                elif op_type == 2:
                    to_set[attr_name] *= attr_val
                elif op_type == 3:
                    to_set[attr_name] /= attr_val
                else:
                    raise NotImplementedError()
            elif op_type == 100:
                # then attrib type is supposed to be string
                if attr_name not in to_set:
                    to_set[attr_name] = ''
                to_set[attr_name] += context.param_value(f'svalue_{i}')
            elif op_type == 200:  # append
                attr_type = context.param_value(f'lvaltype_{i}')
                if attr_name not in to_set:
                    to_set[attr_name] = []
                if attr_type == NodeParameterType.INT.value:
                    to_set[attr_name].append(context.param_value(f'livalue_{i}'))
                elif attr_type == NodeParameterType.FLOAT.value:
                    to_set[attr_name].append(context.param_value(f'lfvalue_{i}'))
                elif attr_type == NodeParameterType.STRING.value:
                    to_set[attr_name].append(context.param_value(f'lsvalue_{i}'))
                else:
                    raise NotImplementedError()
            elif op_type == 201:  # pop
                if attr_name in to_set:
                    to_set[attr_name].pop()
            elif op_type == 202:  # slice
                if attr_name not in to_set:
                    to_set[attr_name] = []
                else:
                    start = context.param_value(f'lsl0value_{i}')
                    end = context.param_value(f'lsl1value_{i}')
                    step = context.param_value(f'lsl2value_{i}')
                    to_set[attr_name] = to_set[attr_name][start:end:step]
            else:
                raise NotImplementedError()

        for attr_name, attr_val in to_set.items():
            res.set_attribute(attr_name, attr_val)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



from pathlib import Path

from typing import Optional


class PluginInfo:
    """
    class for getting information about a plugin

    """
    def __init__(self, file_path, parent_package=None):
        self.__file_path = Path(file_path)

        self.__parent_package = Path(parent_package) if parent_package is not None else None
        self.__parent_package_data = None
        if self.__parent_package is not None:
            self.__parent_package_data = self.__parent_package / 'data'
            if not self.__parent_package_data.exists():
                self.__parent_package_data = None

        self.__package_name = None

    def package_name(self) -> Optional[str]:
        if self.__parent_package is None:
            return None
        if self.__package_name is None:
            self.__package_name = self.__parent_package.name
        return self.__package_name

    def package_root(self) -> Optional[Path]:
        return self.__parent_package

    def package_data(self) -> Optional[Path]:
        return self.__parent_package_data

    def node_definition_file_path(self) -> Path:
        return self.__file_path

    def __str__(self):
        return f'Plugin from {self.node_definition_file_path()}, part of {self.package_name()}'



from ..interfaces import MessageReceiverFactory
from ..messages import Message
from ..address import DirectAddress
from ..exceptions import AddressTypeNotSupportedError
from .tcp_message_receiver import TcpMessageReceiver

from typing import Callable, Awaitable


class TcpMessageReceiverFactory(MessageReceiverFactory):
    def __init__(self, backlog=4096):
        self.__backlog = backlog

    async def create_receiver(self, address: DirectAddress, message_callback: Callable[[Message], Awaitable[bool]]) -> TcpMessageReceiver:
        if address.count(':') != 1:
            raise AddressTypeNotSupportedError(f'address "{address}" is not of "<host>:<port>" format')
        host, sport = address.split(':')
        receiver = TcpMessageReceiver((host, int(sport)), message_callback, socket_backlog=self.__backlog)
        await receiver.start()
        return receiver



from threading import Lock
from contextlib import contextmanager


class RWLock:
    def __init__(self):
        self.__helper_lock = Lock()
        self.__wlock = Lock()
        self.__num_readers = 0

    def acquire_reader(self):
        with self.__helper_lock:
            if self.__num_readers == 0:
                self.__wlock.acquire()
            self.__num_readers += 1

    def release_reader(self):
        with self.__helper_lock:
            self.__num_readers -= 1
            if self.__num_readers == 0:
                self.__wlock.release()

    def acquire_writer(self):
        self.__wlock.acquire()

    def release_writer(self):
        self.__wlock.release()

    @contextmanager
    def reader_lock(self):
        try:
            self.acquire_reader()
            yield
        finally:
            self.release_reader()

    @contextmanager
    def writer_lock(self):
        try:
            self.acquire_writer()
            yield
        finally:
            self.release_writer()



import os
from math import sqrt, floor, ceil
import re

from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.uidata import NodeParameterType
from lifeblood.processingcontext import ProcessingContext
from lifeblood.invocationjob import InvocationJob, InvocationRequirements
from lifeblood.enums import WorkerType

from typing import Iterable


def node_class():
    return FileWatcher


watcher_onmoved = \
'''
    def on_moved(self, event):
        super(FileWatcher, self).on_moved(event)
        ftype = 'dir' if event.is_directory else 'file'
        if ftype not in {ct}:
            return
        if all(not fnmatch(os.path.basename(event.src_path), x) for x in {fpats}):
            return
        lifeblood_connection.create_task(f'{{event.src_path}} moved to {{event.dest_path}}',
                                        {{'src_file': event.src_path,
                                         'file': event.dest_path,
                                         'file_event': 'moved',
                                         'file_type': ftype}})
'''

watcher_oncreated = \
'''
    def on_created(self, event):
        super(FileWatcher, self).on_created(event)
        ftype = 'dir' if event.is_directory else 'file'
        if ftype not in {ct}:
            return
        if all(not fnmatch(os.path.basename(event.src_path), x) for x in {fpats}):
            return
        lifeblood_connection.create_task(f'{{event.src_path}} created',
                                        {{'file': event.src_path,
                                         'file_event': 'created',
                                         'file_type': ftype}})
'''

watcher_ondeleted = \
'''
    def on_deleted(self, event):
        super(FileWatcher, self).on_deleted(event)
        ftype = 'dir' if event.is_directory else 'file'
        if ftype not in {ct}:
            return
        if all(not fnmatch(os.path.basename(event.src_path), x) for x in {fpats}):
            return
        lifeblood_connection.create_task(f'{{event.src_path}} deleted',
                                        {{'file': event.src_path,
                                         'file_event': 'deleted',
                                         'file_type': ftype}})
'''

watcher_onmodified = \
'''
    def on_modified(self, event):
        super(FileWatcher, self).on_modified(event)
        ftype = 'dir' if event.is_directory else 'file'
        if ftype not in {ct}:
            return
        if all(not fnmatch(os.path.basename(event.src_path), x) for x in {fpats}):
            return
        lifeblood_connection.create_task(f'{{event.src_path}} modified',
                                        {{'file': event.src_path,
                                         'file_event': 'modified',
                                         'file_type': ftype}})
'''

watcher_onclosed = \
'''
    def on_closed(self, event):
        super(FileWatcher, self).on_closed(event)
        ftype = 'dir' if event.is_directory else 'file'
        if ftype not in {ct}:
            return
        if all(not fnmatch(os.path.basename(event.src_path), x) for x in {fpats}):
            return
        lifeblood_connection.create_task(f'{{event.src_path}} modified',
                                        {{'file': event.src_path,
                                         'file_event': 'closed',
                                         'file_type': ftype}})
'''

watcher_existings = \
'''
    from pathlib import Path
    
    def _check_one_dir(dir, do_dirs, do_files, recursive):
        for file in dir.iterdir():
            if all(not fnmatch(file.name, x) for x in {fpats}):
                continue
            if file.is_dir():
                if do_dirs:
                    lifeblood_connection.create_task(f'{{file}} {event_type}',
                                                    {{'file': str(file),
                                                     'file_event': '{event_type}',
                                                     'file_type': 'dir'}})
                if recursive:
                    _check_one_dir(file, do_dirs, do_files, recursive)
            elif file.is_file() and do_files:
                lifeblood_connection.create_task(f'{{file}} {event_type}',
                                                {{'file': str(file),
                                                 'file_event': '{event_type}',
                                                 'file_type': 'file'}})

    _check_one_dir(Path(path_to_watch), {do_dirs}, {do_files}, recursive)
'''

watcher_code = \
'''
import os
from fnmatch import fnmatch
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
import lifeblood_connection


class FileWatcher(FileSystemEventHandler):
{onclosed}
{onmoved}
{oncreated}    
{ondeleted}
{onmodified}

def main(path_to_watch, recursive):
    {existings}
    file_observer = Observer()
    file_observer.schedule(FileWatcher(), path_to_watch, recursive)
    file_observer.start()
    file_observer.join()

if __name__ == '__main__':
    import sys
    sys.exit(main("{dir_path}", {recursive}) or 0)

'''


class FileWatcher(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'file watcher'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'file', 'watcher'

    @classmethod
    def type_name(cls) -> str:
        return 'file_watcher'

    def __init__(self, name):
        super(FileWatcher, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_output_for_spawned_tasks()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('path', 'dir/file to watch', NodeParameterType.STRING, '')
                ui.add_parameter('recursive', 'recursive', NodeParameterType.BOOL, False)

            ui.add_parameter('doexistings', None, NodeParameterType.STRING, 'create tasks for existing items:', readonly=True)
            with ui.parameters_on_same_line_block():
                ui.add_parameter('exist dir', 'dirs', NodeParameterType.BOOL, False)
                ui.add_parameter('exist file', 'files', NodeParameterType.BOOL, False)
                ui.add_parameter('exist pattern', 'file pattern', NodeParameterType.STRING, '*')
                ui.add_parameter('exist label', 'label for existing items', NodeParameterType.STRING, 'closed')

            ui.add_separator()

            with ui.parameters_on_same_line_block():
                # ui.add_parameter('onclose', None, NodeParameterType.STRING, 'on close', readonly=True)
                ui.add_parameter('onclose dir', 'dirs/files', NodeParameterType.BOOL, False)
                ui.add_parameter('onclose file', 'pattern', NodeParameterType.BOOL, True)
                ui.add_parameter('onclose pattern', ': on close', NodeParameterType.STRING, '*')

            with ui.parameters_on_same_line_block():
                # ui.add_parameter('oncreate', None, NodeParameterType.STRING, 'on create', readonly=True)
                ui.add_parameter('oncreate dir', 'dirs/files', NodeParameterType.BOOL, False)
                ui.add_parameter('oncreate file', 'pattern', NodeParameterType.BOOL, False)
                ui.add_parameter('oncreate pattern', ': on create', NodeParameterType.STRING, '*')
            with ui.parameters_on_same_line_block():
                # ui.add_parameter('ondelete', None, NodeParameterType.STRING, 'on delete', readonly=True)
                ui.add_parameter('ondelete dir', 'dirs/files', NodeParameterType.BOOL, False)
                ui.add_parameter('ondelete file', 'pattern', NodeParameterType.BOOL, False)
                ui.add_parameter('ondelete pattern', ': on delete', NodeParameterType.STRING, '*')
            with ui.parameters_on_same_line_block():
                # ui.add_parameter('onmove', None, NodeParameterType.STRING, 'on move', readonly=True)
                ui.add_parameter('onmove dir', 'dirs/files', NodeParameterType.BOOL, False)
                ui.add_parameter('onmove file', 'pattern', NodeParameterType.BOOL, False)
                ui.add_parameter('onmove pattern', ': on move', NodeParameterType.STRING, '*')
            with ui.parameters_on_same_line_block():
                # ui.add_parameter('onmodify', None, NodeParameterType.STRING, 'on modify', readonly=True)
                ui.add_parameter('onmodify dir', 'dirs/files', NodeParameterType.BOOL, False)
                ui.add_parameter('onmodify file', 'pattern', NodeParameterType.BOOL, False)
                ui.add_parameter('onmodify pattern', ': on modify', NodeParameterType.STRING, '*')
            ui.add_parameter('note_label', None, NodeParameterType.STRING,
                             'NOTE: Already running tasks will only pick up changes after task restart',
                             readonly=True)

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        inv = InvocationJob(['python', ':/watcher.py'], requirements=InvocationRequirements(worker_type=WorkerType.SCHEDULER_HELPER))
        path = context.param_value('path')
        if not os.path.exists(path):
            raise ProcessingError('watch path must exist')
        if not os.path.isabs(path):
            raise ProcessingError('watch path must be absolute path')
        clo_ct = []
        cre_ct = []
        del_ct = []
        mov_ct = []
        mod_ct = []
        exs_ct = []
        all = {'onclose': clo_ct, 'oncreate': cre_ct, 'ondelete': del_ct, 'onmove': mov_ct, 'onmodify': mod_ct, 'exist': exs_ct}
        for base in ('oncreate', 'ondelete', 'onmove', 'onmodify', 'onclose', 'exist'):
            for ftype in ('dir', 'file'):
                if context.param_value(f'{base} {ftype}'):
                    all[base].append(ftype)
        inv.set_extra_file('watcher.py', watcher_code.format(
            dir_path=path,
            recursive=context.param_value('recursive'),
            onclosed=watcher_onclosed.format(ct=repr(clo_ct),
                                             fpats=repr(tuple(re.split('[, ]+', context.param_value(f'onclose pattern').strip())))
                                             ) if len(clo_ct) > 0 else '',
            onmoved=watcher_onmoved.format(ct=repr(mov_ct),
                                           fpats=repr(tuple(re.split('[, ]+', context.param_value(f'onmove pattern').strip())))
                                           ) if len(mov_ct) > 0 else '',
            oncreated=watcher_oncreated.format(ct=repr(cre_ct),
                                               fpats=repr(tuple(re.split('[, ]+', context.param_value(f'oncreate pattern').strip())))
                                               ) if len(cre_ct) > 0 else '',
            ondeleted=watcher_ondeleted.format(ct=repr(del_ct),
                                               fpats=repr(tuple(re.split('[, ]+', context.param_value(f'ondelete pattern').strip())))
                                               ) if len(del_ct) > 0 else '',
            onmodified=watcher_onmodified.format(ct=repr(mod_ct),
                                                 fpats=repr(tuple(re.split('[, ]+', context.param_value(f'onmodify pattern').strip())))
                                                 ) if len(mod_ct) > 0 else '',
            existings=watcher_existings.format(do_dirs='dir' in exs_ct,
                                               do_files='file' in exs_ct,
                                               event_type=context.param_value('exist label'),
                                               fpats=repr(tuple(re.split('[, ]+', context.param_value('exist pattern').strip())))
                                               ) if len(exs_ct) > 0 else ''
            ))
        #print(inv.extra_files()['watcher.py'])
        return ProcessingResult(inv)

    def postprocess_task(self, context: ProcessingContext) -> ProcessingResult:
        return ProcessingResult()



"""
this module is supposed to be accessible for tasks ran by worker.
this module should be kept minimal, with only standard modules and maximum compatibility
"""
from __future__ import print_function
import os
import time
import errno
import pickle
import json
import threading
import socket
import struct
import uuid
try:
    from typing import Optional, Tuple, Union
except ImportError:
    pass


class MessageSendError(RuntimeError):
    pass


class MessageReceiveTimeout(RuntimeError):
    pass


class InvocationMessageReceiverFuture(threading.Thread):
    def __init__(self, addressee, timeout=None):
        super().__init__()
        self.__addressee = addressee
        self.__timeout = timeout
        self.__result = None
        self.__exception = None

    def run(self) -> None:
        try:
            self.__result = _message_to_invocation_receive_blocking(self.__addressee, self.__timeout)
        except Exception as e:
            self.__exception = e
        _clear_me_from_threads_to_wait()

    def is_done(self):
        return not self.is_alive()

    def wait(self, timeout=None):  # type: (Optional[float]) -> Tuple[int, bytes]
        """
        wait for operation to finish and return result, or raise exception if error happened
        """
        self.join(timeout)
        if self.is_alive():  # timeout happened
            raise TimeoutError()
        if self.__exception:
            raise self.__exception
        return self.__result


def send_string(sock, text):  # type: (socket.socket, str) -> None
    bts = text.encode('UTF-8')
    sock.sendall(struct.pack('>Q', len(bts)))
    sock.sendall(bts)


def recv_string(sock):  # type: (socket.socket) -> str
    data_size, = struct.unpack('>Q', sock.recv(8))
    return sock.recv(data_size).decode('UTF-8')


class EnvironmentResolverArguments:
    """
    this is a copy of envirionment_resolver.EnvironmentResolverArguments class purely for pickling
    """
    def __init__(self, resolver_name=None, arguments=None):
        if arguments is None:
            arguments = {}
        if resolver_name is None and len(arguments) > 0:
            raise ValueError('if name is None - no arguments are allowed')
        self.__resolver_name = resolver_name
        self.__args = arguments


class TaskSpawn:
    """
    this class is a pickle compatible shrunk copy of lifeblood.taskspawn.TaskSpawn
    keep it up-to-date
    and keep it 2-3 compatible!!
    """
    def __init__(self, name, source_invocation_id, env_args=None, task_attributes=None):
        self.__name = name
        self.__attributes = dict(task_attributes or {})
        self.__env_args = env_args
        self.__forced_node_task_id_pair = None
        self.__from_invocation_id = source_invocation_id
        self.__output = 'spawned'
        self._create_as_spawned = True
        self.__extra_groups = []
        self.__default_priority = None

    def create_as_spawned(self):
        return self._create_as_spawned

    def force_set_node_task_id(self, node_id, task_id):
        self.__forced_node_task_id_pair = (node_id, task_id)

    def forced_node_task_id(self):
        return self.__forced_node_task_id_pair

    def source_invocation_id(self):
        return self.__from_invocation_id

    def set_node_output_name(self, new_name):
        self.__output = new_name

    def node_output_name(self):
        return self.__output

    def name(self):
        return self.__name

    def default_priority(self):  # type: () -> Optional[float]
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is

        :return: default priority
        """
        return self.__default_priority

    def add_extra_group_name(self, group_name):
        self.__extra_groups.append(group_name)

    def extra_group_names(self):
        return self.__extra_groups

    def set_default_priority(self, priority):  # type: (float) -> None
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is
        """
        self.__default_priority = priority

    def set_name(self, name):
        self.__name = name

    def set_attribute(self, attr_name, attr_value):
        self.__attributes[attr_name] = attr_value

    def remove_attribute(self, attr_name):
        del self.__attributes[attr_name]

    def attribute_value(self, attr_name):
        return self.__attributes.get(attr_name, None)

    def _attributes(self):
        return self.__attributes

    def environment_arguments(self):
        return self.__env_args

    def serialize(self):
        return pickle.dumps(self)


_threads_to_wait = []
_threads_to_wait_lock = threading.Lock()


def _connect_to_worker(timeout=None):
    addrport = os.environ['LIFEBLOOD_RUNTIME_SCHEDULER_ADDR']
    addr, sport = addrport.rsplit(':', 1)
    port = int(sport)
    if timeout is None:
        sock = socket.create_connection((addr, port))
    else:
        sock = socket.create_connection((addr, port), timeout=timeout)
    # negotiate protocol
    sock.sendall(struct.pack('>QII', 1, 1, 0))
    assert struct.unpack('>II', sock.recv(8)) == (1, 0), 'server does not support our protocol version'
    return sock


def create_task(name, attributes, env_arguments=None, blocking=False):
    """
    creates a new task with name and attributes.
    if env_attributes is None - environment is inherited fully from the parent,
      otherwise env_attributes completely override env from the parent task.

    if blocking is False - create_task creates a thread and exits immediately.
      this is the default as you would usually want to contact scheduler in parallel with actual work.
    """
    invocation_id = int(os.environ['LIFEBLOOD_RUNTIME_IID'])
    spawn = TaskSpawn(name, invocation_id, task_attributes=attributes, env_args=env_arguments)

    def _send():
        sock = _connect_to_worker(timeout=30)

        send_string(sock, 'spawn')

        data = spawn.serialize()
        sock.sendall(struct.pack('>Q', len(data)))
        sock.sendall(data)
        res = sock.recv(13)  # >I?Q  13 should be small enough to ensure receiving in one call
        good, is_not_none, task_id = struct.unpack('>I?Q', res)

        if not blocking:
            _clear_me_from_threads_to_wait()
        return task_id if is_not_none else -1

    if blocking:
        return _send()
    else:
        thread = threading.Thread(target=_send)
        _threads_to_wait.append(thread)
        thread.start()  # and not care


def set_attributes(attribs, blocking=False):  # type: (dict, bool) -> None
    def _send():
        sock = _connect_to_worker(timeout=30)

        send_string(sock, 'tupdateattribs')
        updata = json.dumps(attribs).encode('UTF-8')
        sock.sendall(struct.pack('>QQQ', task_id, len(updata), 0))
        sock.sendall(updata)
        sock.recv(1)  # recv confirmation

        if not blocking:
            _clear_me_from_threads_to_wait()

    task_id = int(os.environ['LIFEBLOOD_RUNTIME_TID'])

    if blocking:
        _send()
    else:
        thread = threading.Thread(target=_send)
        with _threads_to_wait_lock:
            _threads_to_wait.append(thread)
        thread.start()  # and not care


def message_to_invocation_send(invocation_id, addressee, message, addressee_timeout=None):  # type: (int, str, bytes, Optional[float]) -> None
    """
    send a message to invocation_id, addressed to addressee.
    This is useful for designing workflows where:
     - child tasks are reporting processing results to running parent
     - child tasks are synchronizing through parent (useful for distributed sims)
     - other stuff

     NOTE: timeout=None means DEFAULT timeout, not no timeout. disabling timeout is not allowed
    """
    def _send():
        sock = _connect_to_worker()

        send_string(sock, 'sendinvmessage')

        sock.sendall(struct.pack('>QQQf', invocation_id, int(os.environ['LIFEBLOOD_RUNTIME_IID']), len(message), addressee_timeout))
        send_string(sock, addressee)
        sock.sendall(message)
        # now get reply
        reply = recv_string(sock)
        if reply != 'delivered':
            raise MessageSendError(reply)

    if addressee_timeout is None or addressee_timeout <= 0:
        addressee_timeout = 90
    _send()


def message_to_invocation_receive(addressee, timeout=None, blocking=True):  # type: (str, Optional[float], bool) -> Union[Tuple[int, bytes], InvocationMessageReceiverFuture]
    """
    await message from another invocation, addressed to addressee
    If graph is not properly designed so that another invocation may not send a message -
    this function will just block forever

    If blocking is False - it will return a Future-like object that caller can wait for.
    NOTE: this future object actually creates a thread, which may be undesireable
    """
    if blocking:
        return _message_to_invocation_receive_blocking(addressee, timeout)
    future = InvocationMessageReceiverFuture(addressee, timeout)
    future.start()
    with _threads_to_wait_lock:
        _threads_to_wait.append(future)
    return future


def _message_to_invocation_receive_blocking(addressee, timeout=None):  # type: (str, Optional[float]) -> Tuple[int, bytes]
    sock = _connect_to_worker()

    send_string(sock, 'recvinvmessage')

    send_string(sock, addressee)
    target_poll_time = 10.0
    min_poll_time = 0.1
    eps = 0.001
    poll_time = max(min_poll_time, (timeout+eps)/max(1.0, round((timeout+eps)/target_poll_time))) if timeout is not None else target_poll_time
    sock.sendall(struct.pack('>f', poll_time))
    ready = False
    waiting_elapsed = 0
    _last_timestamp = time.time()
    while not ready:
        _last_timestamp = time.time()
        ready, = struct.unpack('>?', sock.recv(1))

        waiting_elapsed += time.time() - _last_timestamp
        _last_timestamp = time.time()
        do_cancel = timeout is not None and waiting_elapsed >= timeout

        if not ready:
            sock.sendall(struct.pack('>?', do_cancel))  # False for no cancel, True for cancel
            if do_cancel:
                raise MessageReceiveTimeout(f'elapsed: {waiting_elapsed}s')
            continue
    source_iid, data_size = struct.unpack('>QQ', sock.recv(16))
    data = sock.recv(data_size)

    return source_iid, data


def _clear_me_from_threads_to_wait():
    with _threads_to_wait_lock:
        me = threading.current_thread()
        try:
            _threads_to_wait.remove(me)
        except ValueError:  # not in the list
            pass


def wait_for_currently_running_async_operations():
    """
    The python program will not end until all non-daemon threads are done.
    So skipping explicit wait for most of the operations is fine

    But sometimes you might need to wait for all current async operations to finish
    and for that you can use this function

    :return:
    """
    global _threads_to_wait
    with _threads_to_wait_lock:
        threads_list_snapshot = _threads_to_wait

    for thread in threads_list_snapshot:
        thread.join()

    with _threads_to_wait_lock:  # cleanup just in case
        for thread in threads_list_snapshot:
            if thread in _threads_to_wait:
                print('WARNING: uncleaned thread!')
                _threads_to_wait.remove(thread)


def get_host_ip():
    """
    returns the ip address of the host running the task.
    if host has multiple interfaces - the one that can connect to the scheduler will be chosen

    :return:
    """
    addrport = os.environ['LIFEBLOOD_RUNTIME_SCHEDULER_ADDR']
    addr, sport = addrport.rsplit(':', 1)

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect((addr, 1))
        myip = s.getsockname()[0]
    except Exception:
        myip = '127.0.0.1'
    finally:
        s.close()
    return myip


def generate_addressee():
    """
    generate random string, nothing more
    """
    return str(uuid.uuid4())


def get_my_invocation_id():
    """
    since this module is run in environment generated by worker,
    there are extra env variables available, including informatino about
    the task+invocation currently being worked on
    """
    return int(os.environ['LIFEBLOOD_RUNTIME_IID'])


def get_free_tcp_port(ip, starting_at=20001):
    """
    returns the next openable port starting at starting_at on the given interface
    NOTE: that port is not reserved, so by the time you try to open it again after this function
    it may be already taken

    :param ip:
    :param starting_at:
    :return:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    for starting_at in range(starting_at, starting_at + 2000):  # big barely sane number of tries, but not infinite
        try:
            s.bind((ip, starting_at))
            s.close()
            break
        except OSError as e:
            if e.errno == errno.EADDRINUSE:
                continue
            raise
    return starting_at



import time
import json
from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.taskspawn import TaskSpawn
from lifeblood.exceptions import NodeNotReadyToProcess
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi
from lifeblood.processingcontext import ProcessingContext

from threading import Lock

from typing import Dict, TypedDict, Set, Iterable, Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from lifeblood.scheduler import Scheduler


class SplitAwaiting(TypedDict):
    arrived: Dict[int, dict]  # num in split -2-> attributes
    awaiting: Set[int]
    first_to_arrive: Optional[int]


def node_class():
    return SplitAwaiterNode


class SplitAwaiterNode(BaseNode):

    @classmethod
    def label(cls) -> str:
        return 'split gatherer'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'split', 'wait', 'synchronization', 'barrier', 'gather', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'split_waiter'

    def __init__(self, name: str):
        super(SplitAwaiterNode, self).__init__(name)
        self.__cache: Dict[int: SplitAwaiting] = {}
        self.__main_lock = Lock()
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('wait for all', 'wait for all', NodeParameterType.BOOL, True)
            with ui.multigroup_parameter_block('transfer_attribs'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('src_attr_name', 'attribute', NodeParameterType.STRING, 'attr1')
                    ui.add_parameter('transfer_type', 'as', NodeParameterType.STRING, 'extend')\
                        .add_menu((('Extend', 'extend'),
                                   ('Append', 'append'),
                                   ('First', 'first'),
                                   ('Sum', 'sum')))
                    ui.add_parameter('dst_attr_name', 'sort by', NodeParameterType.STRING, 'attr1')
                    ui.add_parameter('sort_by', None, NodeParameterType.STRING, '_builtin_id')
                    ui.add_parameter('reversed', 'reversed', NodeParameterType.BOOL, False)

    def ready_to_process_task(self, task_dict) -> bool:
        context = ProcessingContext(self, task_dict)
        split_id = context.task_field('split_id')
        # we don't even need to lock
        return split_id not in self.__cache or \
               not context.param_value('wait for all') or \
               context.task_field('split_element') not in self.__cache[split_id]['arrived'] or \
               self.__cache[split_id]['arrived'].keys() == self.__cache[split_id]['awaiting']

    def process_task(self, context) -> ProcessingResult: #TODO: not finished, attrib not taken into account, rethink return type
        orig_id = context.task_field('split_origin_task_id')
        split_id = context.task_field('split_id')
        task_id = context.task_field('id')
        if orig_id is None:  # means no splits - just pass through
            return ProcessingResult()
        with self.__main_lock:
            if split_id not in self.__cache:
                self.__cache[split_id] = {'arrived': {},
                                          'awaiting': set(range(context.task_field('split_count'))),
                                          'processed': set(),
                                          'first_to_arrive': None}
            if self.__cache[split_id]['first_to_arrive'] is None and len(self.__cache[split_id]['arrived']) == 0:
                self.__cache[split_id]['first_to_arrive'] = task_id
            if context.task_field('split_element') not in self.__cache[split_id]['arrived']:
                self.__cache[split_id]['arrived'][context.task_field('split_element')] = json.loads(context.task_field('attributes'))
                self.__cache[split_id]['arrived'][context.task_field('split_element')]['_builtin_id'] = task_id

        # we will not wait in loop or we risk deadlocking threadpool
        # check if everyone is ready
        changed = False
        try:
            if context.param_value('wait for all'):
                with self.__main_lock:
                    if self.__cache[split_id]['arrived'].keys() == self.__cache[split_id]['awaiting']:
                        res = ProcessingResult()
                        res.kill_task()
                        self.__cache[split_id]['processed'].add(context.task_field('split_element'))
                        attribs_to_promote = {}
                        if self.__cache[split_id]['first_to_arrive'] == task_id:
                            # transfer attributes  # TODO: delete cache for already processed splits
                            num_attribs = context.param_value('transfer_attribs')
                            for i in range(num_attribs):
                                src_attr_name = context.param_value(f'src_attr_name_{i}')
                                transfer_type = context.param_value(f'transfer_type_{i}')
                                dst_attr_name = context.param_value(f'dst_attr_name_{i}')
                                sort_attr_name = context.param_value(f'sort_by_{i}')
                                sort_reversed = context.param_value(f'reversed_{i}')
                                if transfer_type == 'append':
                                    gathered_values = []
                                    for attribs in sorted(self.__cache[split_id]['arrived'].values(), key=lambda x: x.get(sort_attr_name, 0), reverse=sort_reversed):
                                        if src_attr_name not in attribs:
                                            continue

                                        attr_val = attribs[src_attr_name]
                                        gathered_values.append(attr_val)
                                    attribs_to_promote[dst_attr_name] = gathered_values
                                elif transfer_type == 'extend':
                                    gathered_values = []
                                    for attribs in sorted(self.__cache[split_id]['arrived'].values(), key=lambda x: x.get(sort_attr_name, 0), reverse=sort_reversed):
                                        if src_attr_name not in attribs:
                                            continue

                                        attr_val = attribs[src_attr_name]
                                        if isinstance(attr_val, list):
                                            gathered_values.extend(attr_val)
                                        else:
                                            gathered_values.append(attr_val)
                                    attribs_to_promote[dst_attr_name] = gathered_values
                                elif transfer_type == 'first':
                                    _acd = self.__cache[split_id]['arrived']
                                    if len(_acd) > 0:
                                        if sort_reversed:
                                            attribs = max(_acd.values(), key=lambda x: x.get(sort_attr_name, 0))
                                        else:
                                            attribs = min(_acd.values(), key=lambda x: x.get(sort_attr_name, 0))
                                        if src_attr_name in attribs:
                                            attribs_to_promote[dst_attr_name] = attribs[src_attr_name]
                                elif transfer_type == 'sum':
                                    # we don't care about the order, assume sum is associative
                                    gathered_values = None
                                    for attribs in self.__cache[split_id]['arrived'].values():
                                        if src_attr_name not in attribs:
                                            continue
                                        if gathered_values is None:
                                            gathered_values = attribs[src_attr_name]
                                        else:
                                            gathered_values += attribs[src_attr_name]
                                    attribs_to_promote[dst_attr_name] = gathered_values
                                else:
                                    raise NotImplementedError(f'transfer type "{transfer_type}" is not implemented')

                            res.remove_split(attributes_to_set=attribs_to_promote)
                        changed = True
                        return res
            else:
                with self.__main_lock:
                    res = ProcessingResult()
                    res.kill_task()
                    self.__cache[split_id]['processed'].add(context.task_field('split_element'))
                    if self.__cache[split_id]['first_to_arrive'] == task_id:
                        res.remove_split()
                    changed = True
                    return res

        finally:
            if self.__cache[split_id]['processed'] == self.__cache[split_id]['awaiting']:  # kinda precheck, to avoid extra lockings
                with self.__main_lock:
                    if self.__cache[split_id]['processed'] == self.__cache[split_id]['awaiting']:  # and proper check inside lock
                        del self.__cache[split_id]
            # if changed:
            #     self._state_changed()  # this cannot be called from non asyncio thread as this.

        raise NodeNotReadyToProcess()

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()

    def __getstate__(self):
        d = super(SplitAwaiterNode, self).__getstate__()
        assert '_SplitAwaiterNode__main_lock' in d
        del d['_SplitAwaiterNode__main_lock']
        return d



import asyncio
from .message_protocol import MessageProtocol, IProtocolInstanceCounter
from ..interfaces import MessageReceiver, MessageStreamFactory
from ..messages import Message
from ..address import DirectAddress
from ..logging import get_logger

from typing import Awaitable, Callable, List, Tuple


class TcpMessageReceiver(MessageReceiver, IProtocolInstanceCounter):
    def __init__(self, address: Tuple[str, int],
                 message_received_callback: Callable[[Message], Awaitable[bool]],
                 *, socket_backlog=4096):
        super().__init__(DirectAddress.from_host_port(*address))
        self.__address = address
        self.__socket_backlog = socket_backlog
        self.__message_received_callback = message_received_callback
        self.__server = None
        self.__active_protocol_instances_count = 0
        self.__protocol_instance_stash: List[MessageProtocol] = []
        self.__no_more_instances = asyncio.Event()
        self.__no_more_instances.set()
        self.__logger = get_logger(f'{self.__class__.__name__}')

    def _protocol_inc_count(self, instance):
        if self.__active_protocol_instances_count == 0:
            self.__no_more_instances.clear()
        self.__active_protocol_instances_count += 1
        self.__protocol_instance_stash.append(instance)

    def _protocol_dec_count(self, instance):
        if self.__active_protocol_instances_count == 0:
            raise RuntimeError('inconsistent protocol instance count!')
        self.__active_protocol_instances_count -= 1
        if self.__active_protocol_instances_count == 0:
            self.__no_more_instances.set()
        self.__protocol_instance_stash.remove(instance)

    def _allowed_new_instances(self) -> bool:
        return self.__server.is_serving()

    async def start(self):
        self.__server = await asyncio.get_event_loop().create_server(lambda: MessageProtocol(self,
                                                                                             self.__address,
                                                                                             self.__message_received_callback),
                                                                     self.__address[0],
                                                                     self.__address[1],
                                                                     backlog=self.__socket_backlog)

    def stop(self):
        self.__logger.debug('stopping tcp message receiver')
        self.__server.close()
        for instance in self.__protocol_instance_stash:
            instance.stop()

    async def wait_till_stopped(self):
        await self.__server.wait_closed()
        await self.__no_more_instances.wait()



import os
from pathlib import Path
import copy
import toml
import asyncio
import tempfile
from threading import Lock
from . import paths
from .logging import get_logger

from typing import Any, List, Tuple, Union, Callable, Set


_conf_cache = {}
_glock = Lock()


def get_config(subname: str) -> "Config":
    global _glock, _conf_cache
    with _glock:
        if subname not in _conf_cache:
            _conf_cache[subname] = Config(subname)
        return _conf_cache[subname]


def set_config_overrides(subname: str, overrides=None):
    """
    convenient method to set config's overrides without actually creating config object explicitly

    :param subname:
    :param overrides:
    :return:
    """
    global _glock, _conf_cache
    with _glock:
        if subname not in _conf_cache:
            _conf_cache[subname] = Config(subname, overrides)
        else:
            _conf_cache[subname].set_overrides(overrides)


def create_default_user_config_file(subname: str, default_config: Union[str, dict], force: bool = False, toml_encoder=None):
    """
    create user configuration file
    useful for initialization, but can be forced to overwrite

    :param subname:
    :param default_config:
    :param force: if true - user config will be overriden even if it exists
    :param toml_encoder: config backend is currently TOML, and i already kinda regret it... so you can pass special encoders here
    :return:
    """
    user_conf_path = paths.config_path('config.toml', subname)
    if os.path.exists(user_conf_path) and not force:
        return
    os.makedirs(os.path.dirname(user_conf_path), exist_ok=True)
    with open(user_conf_path, 'w') as f:
        if isinstance(default_config, str):
            f.write(default_config)
        else:
            toml.dump(default_config, f, encoder=toml_encoder)


def get_local_scratch_path():
    return os.path.join(tempfile.gettempdir(), 'lifeblood', 'shared')


class Config:
    __logger = get_logger('config')

    class OverrideNotFound(RuntimeError):
        pass

    def __init__(self, subname: str, base_name: str = 'config', overrides=None):
        config_path = paths.config_path(f'{base_name}.toml', subname)
        configd_path = paths.config_path(f'{base_name}.d', subname)
        self.__writable_config_path = config_path
        self.__conf_lock = Lock()
        self.__write_file_lock = Lock()

        self.__sources: List["Path"] = []
        self.__broken_sources: Set["Path"] = set()

        self.__config_paths_to_check = [config_path]
        if configd_path.exists() and configd_path.is_dir():
            self.__config_paths_to_check.extend(configd_path.iterdir())

        self.__stuff = {}

        self.__encoder_generator = None
        self.__overrides = {}
        self.set_overrides(overrides)

        self.reload(keep_overrides=True)

    @classmethod
    def __update_dicts(cls, main: dict, secondary: dict):
        for key, value in secondary.items():
            if isinstance(value, dict) and isinstance(main.get(key), dict):
                cls.__update_dicts(main[key], value)
                continue
            main[key] = value

    def reload(self, keep_overrides=True) -> None:
        self.__stuff = {}
        if not keep_overrides:
            self.__overrides = {}

        paths_to_check = self.__config_paths_to_check
        # append writable path only for the reload time
        # we don't want to mix paths added to the list from legal sources and this user overridable one
        if self.__writable_config_path not in paths_to_check:
            paths_to_check.append(self.__writable_config_path)

        for config_path in paths_to_check:
            if not config_path.exists():
                continue
            try:
                with config_path.open('r') as f:
                    self.__update_dicts(self.__stuff, toml.load(f))
            except Exception as e:
                self.__logger.error(f'failed to load config file {config_path}, skipping')
                self.__broken_sources.add(config_path)
            else:
                self.__sources.append(config_path)

    def writeable_file(self) -> "Path":
        """
        Get the path to the file chis config changes will be saved into.
        The file might not yet exist

        :return:
        """
        return self.__writable_config_path

    def override_config_save_location(self, path: Union[Path, str]):
        """
        change default config write location

        :return:
        """
        self.__writable_config_path = Path(path)

    def loaded_files(self) -> Tuple["Path"]:
        """
        list files from which this config was sourced

        :return: tuple of file paths
        """
        return tuple(self.__sources)

    def broken_files(self) -> Set["Path"]:
        """
        returns set of paths that look like configs, but had errors when parsing
        """
        return set(self.__broken_sources)

    def set_overrides(self, overrides: dict) -> None:
        """
        sets overrides to a prepared dictionary of items.
        :param overrides:
        """
        if overrides is not None:
            self.__overrides = copy.deepcopy(overrides)
        else:
            self.__overrides = {}

    def set_override(self, option_name: str, val: Any) -> None:
        """
        set one item override
        :param option_name: option path, like foo.bar.cat.dog
        :param val: any serializable value
        :return:
        """
        names = option_name.split('.')
        clevel = self.__overrides
        for name in names[:-1]:
            if name not in clevel:
                clevel[name] = {}
            clevel = clevel[name]
        clevel[names[-1]] = val

    async def get_option(self, option_name: str, default_val: Any = None) -> Any:
        return await asyncio.get_event_loop().run_in_executor(None, self.get_option_noasync, option_name, default_val)

    def _get_option_in_overrides(self, option_name: str):
        names = option_name.split('.')
        clevel = self.__overrides
        for name in names:
            if name not in clevel:
                raise Config.OverrideNotFound()
            clevel = clevel[name]
        return clevel

    def has_option_noasync(self, option_name):
        class _SomethingStrangeL:
            pass
        return self.get_option_noasync(option_name, default_val=_SomethingStrangeL) is not _SomethingStrangeL

    @staticmethod
    def _split_config_names(option_name: str) -> Tuple[str, ...]:
        names = []
        in_quotes = False
        mark = 0
        word_parts = []
        for i, l in enumerate(option_name):
            if l == '.' and not in_quotes:
                word_parts.append(option_name[mark:i])
                names.append(''.join(word_parts))
                if names[-1] == '':
                    raise ValueError(f'"{option_name}" is not a valid option_name')
                word_parts.clear()
                mark = i + 1
            elif l == '"':
                word_parts.append(option_name[mark:i])
                mark = i + 1
                in_quotes = not in_quotes
        if in_quotes:
            raise ValueError(f'"{option_name}" is not a valid option_name')
        word_parts.append(option_name[mark:])
        names.append(''.join(word_parts))
        if names[-1] == '':
            raise ValueError(f'"{option_name}" is not a valid option_name')
        return tuple(names)


    def get_option_noasync(self, option_name: str, default_val: Any = None) -> Any:
        try:
            return self._get_option_in_overrides(option_name)
        except Config.OverrideNotFound:
            pass
        with self.__conf_lock:  # to prevent config corruption when running in parallel in executor
            names = self._split_config_names(option_name)
            clevel = self.__stuff
            for name in names:
                if name not in clevel:
                    return default_val
                clevel = clevel[name]

            return clevel  # TODO: return IMMUTABLE shit! this can be a list or a dict, and it can be nested too!

    def _set_option_noasync_nolock(self, option_name: str, value) -> None:
        names = self._split_config_names(option_name)
        clevel = self.__stuff
        for name in names:
            last = name == names[-1]
            if name not in clevel and not last:
                clevel[name] = {}
            if last:
                clevel[name] = value
                break
            clevel = clevel[name]
        self.write_config_noasync()

    def set_option_noasync(self, option_name: str, value: Any) -> None:
        with self.__conf_lock:
            self._set_option_noasync_nolock(option_name, value)

    async def set_option(self, option_name: str, value: Any) -> None:
        return await asyncio.get_event_loop().run_in_executor(None, self.set_option_noasync, option_name, value)

    def set_toml_encoder_generator(self, generator: Callable):
        self.__encoder_generator = generator

    def write_config_noasync(self):
        with self.__write_file_lock:
            if not os.path.exists(self.__writable_config_path):
                os.makedirs(os.path.dirname(self.__writable_config_path), exist_ok=True)
            with open(self.__writable_config_path, 'w') as f:
                toml.dump(self.__stuff, f, encoder=self.__encoder_generator() if self.__encoder_generator is not None else None)

    async def write_config(self):
        return await asyncio.get_event_loop().run_in_executor(None, self.write_config_noasync)



import asyncio
import struct
from .logging import get_logger
from .enums import WorkerState

from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from .simple_worker_pool import WorkerPool


class WorkerPoolProtocol(asyncio.StreamReaderProtocol):
    def __init__(self, worker_pool: "WorkerPool", limit=2 ** 16, logger=None):
        self.__logger = logger or get_logger(self.__class__.__name__.lower())
        self.__timeout = 60
        self.__worker_pool = worker_pool
        self.__reader = asyncio.StreamReader(limit=limit)
        self.__saved_references = []
        super(WorkerPoolProtocol, self).__init__(self.__reader, self.connection_cb)

    async def connection_cb(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        try:
            prot = await asyncio.wait_for(reader.readexactly(4), self.__timeout)
            if prot != b'\0\0\0\0':
                raise NotImplementedError()

            worker_id = struct.unpack('>Q', await asyncio.wait_for(reader.readexactly(8), self.__timeout))[0]

            while True:
                command = await asyncio.wait_for(reader.readline(), timeout=self.__timeout)  # type: bytes
                if command.endswith(b'\n'):
                    command = command[:-1]
                self.__logger.debug(f'got command {command}')
                if command == b'state_report':
                    state = WorkerState(struct.unpack('>Q', await reader.readexactly(8))[0])
                    await self.__worker_pool._worker_state_change(worker_id, state)
                    writer.write(b'\1')
                elif reader.at_eof():
                    self.__logger.debug('connection closed')
                    return
                else:
                    raise NotImplementedError(f'{command} command is not implemented')
                await writer.drain()

        except asyncio.exceptions.TimeoutError as e:
            self.__logger.error('connection timeout')
        except ConnectionResetError as e:
            self.__logger.exception('connection was reset. disconnected %s', e)
        except ConnectionError as e:
            self.__logger.exception('connection error. disconnected %s', e)
        except Exception as e:
            self.__logger.exception('unknown error. disconnected %s', e)
            raise
        finally:
            writer.close()
            await writer.wait_closed()
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())

    # def connection_lost(self, exc):
    #     self.__
    #     super(WorkerPoolProtocol, self).connection_lost(exc)


class WorkerPoolClient:
    async def write_string(self, s: str):
        b = s.encode('UTF-8')
        self.__writer.write(struct.pack('>Q', len(b)))
        self.__writer.write(b)

    def __init__(self, ip: str, port: int, worker_id: int):
        self.__logger = get_logger('workerpoolclient')
        self.__conn_task = asyncio.create_task(asyncio.open_connection(ip, port))
        self.__worker_id = worker_id
        self.__reader: Optional[asyncio.StreamReader] = None
        self.__writer: Optional[asyncio.StreamWriter] = None

    async def __aenter__(self) -> "WorkerPoolClient":
        await self._ensure_conn_open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        self.__writer.close()
        await self.__writer.wait_closed()

    async def _ensure_conn_open(self):
        if self.__reader is not None:
            return
        self.__reader, self.__writer = await self.__conn_task
        self.__writer.write(b'\0\0\0\0')
        self.__writer.write(struct.pack('>Q', self.__worker_id))

    async def report_state(self, state: WorkerState):
        self.__writer.write(b'state_report\n')
        self.__writer.write(struct.pack('>Q', state.value))
        await self.__writer.drain()
        assert await self.__reader.readexactly(1) == b'\1'



import platform


if platform.system().lower() == 'windows':
    import socket
    # now before you go all crazy about the line below - we ONLY use SO_REUSEPORT in broadcasting
    # otherwise it shouldn't break anything....
    # but yeah....
    socket.SO_REUSEPORT = socket.SO_REUSEADDR



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionSfMTransform


class AlicevisionSfMTransform(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionSfMTransform, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Structure From Motion', NodeParameterType.STRING, "`task['av_structure_from_motion']`")
            ui.add_separator()
            ui.add_parameter('output', 'Output Structure From Motion', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/SfMTransform/sfm.abc")
            ui.add_parameter('outputViewsAndPoses', 'Output Views And Poses', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/SfMTransform/cameras.sfm")
            ui.add_separator()
            # TODO: add modes: transformation, manual, from_single_camera
            ui.add_parameter('method', 'Transformation Method', NodeParameterType.STRING, 'auto_from_cameras')\
                .add_menu((('auto from cameras', 'auto_from_cameras'),
                           ('auto from landmarks', 'auto_from_landmarks'),
                           ('from center camera', 'from_center_camera'),
                           ('from markers', 'from_markers')))
            ui.add_parameter('landmarksDescriberTypes', 'Describer Types', NodeParameterType.STRING, 'sift,dspsift,akaze')  # TODO: make a set of checkboxes
            ui.add_parameter('scale', 'Additional Scale', NodeParameterType.FLOAT, 1.0)
            # TODO: missing markers parameter
            ui.add_parameter('applyScale', 'Apply Scale', NodeParameterType.BOOL, True)
            ui.add_parameter('applyRotation', 'Apply Rotation', NodeParameterType.BOOL, True)
            ui.add_parameter('applyTranslation', 'Apply Translation', NodeParameterType.BOOL, True)

    @classmethod
    def label(cls) -> str:
        return 'AV SfM Transform'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'sfm', 'structure', 'motion', 'transform'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_utils_sfm_transform'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')
        output_vap = Path(context.param_value('outputViewsAndPoses'))
        if not output_vap.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.parent.mkdir(parents=True, exist_ok=True)
        output_vap.parent.mkdir(parents=True, exist_ok=True)

        args = ['aliceVision_utils_sfmTransform', '--verboseLevel', 'info']
        args += ['--input', context.param_value('input').strip()]

        for param_name in ('method', 'landmarksDescriberTypes', 'scale',
                           'applyScale', 'applyRotation', 'applyTranslation'):
            args += [f'--{param_name}', context.param_value(param_name)]

        args += ['--output', output]
        args += ['--outputViewsAndPoses', output_vap]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_structure_from_motion', str(output))
        res.set_attribute('av_viewes_and_poses', str(output_vap))
        return res



from enum import Enum


class NodeParameterType(Enum):
    INT = 0
    BOOL = 1
    FLOAT = 2
    STRING = 3


class SchedulerMode(Enum):
    STANDARD = 0
    DORMANT = 1


class WorkerState(Enum):
    OFF = 0
    IDLE = 1
    BUSY = 2
    ERROR = 3
    INVOKING = 4  # worker is being fed data to start an invocation job
    UNKNOWN = 5  # this state the worker should be initialized in, then scheduler should figure it out


class WorkerPingState(Enum):
    OFF = 0
    CHECKING = 2
    ERROR = 3
    WORKING = 1
    UNKNOWN = 5


class TaskScheduleStatus(Enum):
    SUCCESS = 0
    FAILED = 1
    BUSY = 2
    EMPTY = 3


class TaskExecutionStatus(Enum):
    FINISHED = 0
    RUNNING = 1


class WorkerPingReply(Enum):
    IDLE = 0
    BUSY = 1


class TaskState(Enum):
    WAITING = 0  # arrived at node, does not know what to do
    GENERATING = 1  # node is generating work load
    READY = 2  # ready to be scheduled
    INVOKING = 11  # task is being switched to IN_PROGRESS
    IN_PROGRESS = 3  # is being worked on by a worker
    POST_WAITING = 4  # task is waiting to be post processed by node
    POST_GENERATING = 5  # task is being post processed by node
    DONE = 6  # done, needs further processing
    ERROR = 7  # some internal error, not allowing to process task. NOT INVOCATION ERROR
    SPAWNED = 8  # spawned tasks are just passed down from node's "spawned" output
    DEAD = 9  # task will not be processed any more
    SPLITTED = 10  # task has been splitted, and will remain idle until splits are gathered


class TaskGroupArchivedState(Enum):
    NOT_ARCHIVED = 0
    ARCHIVED = 1


class WorkerType(Enum):
    STANDARD = 0
    SCHEDULER_HELPER = 1


class InvocationState(Enum):
    IN_PROGRESS = 0
    FINISHED = 1
    INVOKING = 2


class SpawnStatus(Enum):
    SUCCEEDED = 0
    FAILED = 1


class ProcessPriorityAdjustment(Enum):
    NO_CHANGE = 0
    LOWER = 1


class UIEventType(Enum):
    UPDATE = 0
    DELETE = 1
    FULL_STATE = 2  # represents a checkpoint in the log that represents the full state, so events before are irrelevant


class InvocationMessageResult(Enum):  # Note: values are picked in a way they are parseable by lifeblood_connection
    DELIVERED = 'delivered'
    ERROR_BAD_IID = 'error-wrong-invid'
    ERROR_RECEIVER_TIMEOUT = 'error-receiver-timeout'
    ERROR_DELIVERY_TIMEOUT = 'error-delivery-timeout'
    ERROR_UNEXPECTED = 'error-unknown'
    ERROR_IID_NOT_RUNNING = 'error-invoc-not-running'
    ERROR_TRANSFER_ERROR = 'error-transfer'



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.enums import NodeParameterType
from lifeblood.taskspawn import TaskSpawn
from lifeblood.text import match_pattern

from typing import Iterable


def node_class():
    return SpawnChildren


class SpawnChildren(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'spawn children'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'spawn', 'child', 'children', 'born', 'birth', 'create'

    @classmethod
    def type_name(cls) -> str:
        return 'spawn_children'

    def __init__(self, name: str):
        super(SpawnChildren, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_output('spawned')
            ui.add_parameter('count', 'spawn this many children', NodeParameterType.INT, 0)
            ui.add_parameter('number attribute', 'child number attribute to create', NodeParameterType.STRING, '')
            ui.add_parameter('child basename', 'children base name', NodeParameterType.STRING, "`task.name`_child")
            ui.add_parameter('inherit attributes', 'attribute pattern to inherit', NodeParameterType.STRING, '*')

    def process_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        basename = context.param_value('child basename')
        pattern = context.param_value('inherit attributes')
        new_attributes_base = {attr: val for attr, val in context.task_attributes().items() if match_pattern(pattern, attr)}
        num_attr_name = context.param_value('number attribute')
        for i in range(context.param_value('count')):
            new_attributes = new_attributes_base.copy()
            if num_attr_name:
                new_attributes[num_attr_name] = i
            newtask = TaskSpawn(f'{basename}{i}', None, task_attributes=new_attributes)
            newtask.force_set_node_task_id(self.id(), context.task_id())
            res.add_spawned_task(newtask)

        return res



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.enums import NodeParameterType

from typing import Iterable


def node_class():
    return DeleteAttributes


class DeleteAttributes(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'delete attributes'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'delete', 'attribute', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'delete_attrib'

    def __init__(self, name: str):
        super(DeleteAttributes, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            with ui.multigroup_parameter_block('num'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('name', 'delete', NodeParameterType.STRING, '')

    def process_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        attrs = set(context.task_attributes().keys())
        for i in range(context.param_value('num')):
            attr_name = context.param_value(f'name_{i}')
            if attr_name == '' or attr_name not in attrs:
                continue
            attrs.remove(attr_name)
            res.remove_attribute(attr_name)
        return res



import asyncio
from io import BufferedIOBase, BytesIO
from .buffered_connection import BufferedReader


class IBufferSerializable:
    def serialize(self, stream: BufferedIOBase):
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, stream: BufferedReader):
        raise NotImplementedError()

    async def serialize_to_streamwriter(self, stream: asyncio.StreamWriter):
        buff = BytesIO()
        await asyncio.get_event_loop().run_in_executor(None, self.serialize, buff)
        stream.write(buff.getbuffer())
